import abjad
import auxjad


def test_simplified_time_signature_ratio_01():
    ratio = auxjad.simplified_time_signature_ratio((4, 8))
    time_signature = abjad.TimeSignature(ratio)
    assert format(time_signature) == abjad.String.normalize(
        r'''abjad.TimeSignature((2, 4))''')


def test_simplified_time_signature_ratio_02():
    ratio = auxjad.simplified_time_signature_ratio((7, 8))
    time_signature = abjad.TimeSignature(ratio)
    assert format(time_signature) == abjad.String.normalize(
        r'''abjad.TimeSignature((7, 8))''')


def test_simplified_time_signature_ratio_03():
    ratio = auxjad.simplified_time_signature_ratio((4, 8), min_denominator=8)
    time_signature = abjad.TimeSignature(ratio)
    assert format(time_signature) == abjad.String.normalize(
        r'''abjad.TimeSignature((4, 8))''')


def test_simplified_time_signature_ratio_04():
    ratio = auxjad.simplified_time_signature_ratio((4, 8), min_denominator=2)
    time_signature = abjad.TimeSignature(ratio)
    assert format(time_signature) == abjad.String.normalize(
        r'''abjad.TimeSignature((1, 2))''')


def test_simplified_time_signature_ratio_05():
    ratio = auxjad.simplified_time_signature_ratio((1, 1))
    time_signature = abjad.TimeSignature(ratio)
    assert format(time_signature) == abjad.String.normalize(
        r'''abjad.TimeSignature((4, 4))''')


def test_simplified_time_signature_ratio_06():
    ratio = auxjad.simplified_time_signature_ratio((1, 1), min_denominator=1)
    time_signature = abjad.TimeSignature(ratio)
    assert format(time_signature) == abjad.String.normalize(
        r'''abjad.TimeSignature((1, 1))''')



import abjad
import pytest
import auxjad


def test_close_container_01():
    container1 = abjad.Container(r"c'4 d'4 e'4 f'4")
    container2 = abjad.Container(r"c'4 d'4 e'4")
    container3 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4")
    container4 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4 d'4 e'4 f'4")
    auxjad.close_container(container1)
    auxjad.close_container(container2)
    auxjad.close_container(container3)
    auxjad.close_container(container4)
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            c'4
            d'4
            e'4
            f'4
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 3/4 %%%
            c'4
            d'4
            e'4
        }
        ''')
    assert format(container3) == abjad.String.normalize(
        r'''
        {
            c'4
            d'4
            e'4
            f'4
            %%% \time 1/4 %%%
            c'4
        }
        ''')
    assert format(container4) == abjad.String.normalize(
        r'''
        {
            c'4
            d'4
            e'4
            f'4
            c'4
            d'4
            e'4
            f'4
        }
        ''')


def test_close_container_02():
    container1 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4 g'")
    container2 = abjad.Container(r"\time 3/4 a2. \time 2/4 c'4")
    container3 = abjad.Container(r"\time 5/4 g1 ~ g4 \time 4/4 af'2")
    auxjad.close_container(container1)
    auxjad.close_container(container2)
    auxjad.close_container(container3)
    assert format(container1) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'4
            d'4
            e'4
            f'4
            %%% \time 1/4 %%%
            g'4
        }
        ''')
    assert format(container2) == abjad.String.normalize(
        r'''
        {
            %%% \time 3/4 %%%
            a2.
            %%% \time 1/4 %%%
            c'4
        }
        ''')
    assert format(container3) == abjad.String.normalize(
        r'''
        {
            %%% \time 5/4 %%%
            g1
            ~
            g4
            %%% \time 2/4 %%%
            af'2
        }
        ''')


def test_close_container_03():
    container = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4 g'")
    auxjad.close_container(container)
    assert format(container) == abjad.String.normalize(
        r'''
        {
            %%% \time 4/4 %%%
            c'4
            d'4
            e'4
            f'4
            %%% \time 1/4 %%%
            g'4
        }
        ''')
    staff = abjad.Staff([container])
    assert format(container) == abjad.String.normalize(
        r'''
        {
            \time 4/4
            c'4
            d'4
            e'4
            f'4
            \time 1/4
            g'4
        }
        ''')


def test_close_container_04():
    container = abjad.Container(r"c'4 d'4 e'4 f'4 g'4")
    time_signature = abjad.TimeSignature((3, 4), partial=(1, 4))
    abjad.attach(time_signature, container[0])
    auxjad.close_container(container)
    assert format(container) == abjad.String.normalize(
        r'''
        {
            %%% \partial 4 %%%
            %%% \time 3/4 %%%
            c'4
            d'4
            e'4
            f'4
            %%% \time 1/4 %%%
            g'4
        }
        ''')


def test_close_container_05():
    container = abjad.Container(r"\time 5/4 g''1 \time 4/4 f'1")
    with pytest.raises(ValueError):
        assert auxjad.close_container(container)



import abjad
import auxjad


def test_are_leaves_tieable_01():
    note1 = abjad.Note(r"c'4")
    note2 = abjad.Note(r"c'4")
    assert auxjad.are_leaves_tieable(note1, note2)


def test_are_leaves_tieable_02():
    note1 = abjad.Note(r"c'2.")
    note2 = abjad.Note(r"c'16")
    note3 = abjad.Note(r"f'''16")
    assert auxjad.are_leaves_tieable(note1, note2)
    assert not auxjad.are_leaves_tieable(note1, note3)
    assert not auxjad.are_leaves_tieable(note2, note3)


def test_are_leaves_tieable_03():
    chord1 = abjad.Chord(r"<c' e' g'>4")
    chord2 = abjad.Chord(r"<c' e' g'>16")
    chord3 = abjad.Chord(r"<f''' fs'''>16")
    assert auxjad.are_leaves_tieable(chord1, chord2)
    assert not auxjad.are_leaves_tieable(chord1, chord3)
    assert not auxjad.are_leaves_tieable(chord2, chord3)


def test_are_leaves_tieable_04():
    container = abjad.Container(r"r4 <c' e'>4 <c' e'>2")
    assert auxjad.are_leaves_tieable(container[1], container[2])


def test_are_leaves_tieable_05():
    container = abjad.Container(r"r4 g'4 r2")
    assert not auxjad.are_leaves_tieable(container[0], container[2])



import abjad


def simplified_time_signature_ratio(ratio: (tuple,
                                            abjad.Duration,
                                            abjad.TimeSignature,
                                            ),
                                    *,
                                    min_denominator: int = 4,
                                    ) -> tuple:
    r"""A function simplifies the ratio of a given time signature respecting a
    minimum denominator value. Input is a tuple of two integers.

    ..  container:: example

        By default, the function simplifies the ratio of numerator/denominator
        using a minimum denominator value of 4 (that is, the denominator will
        not get smaller than 4). In the case below, (2, 4) is the simplest
        representation of the ratio (4, 8) with a denominator equal to or
        larger than 4.

        >>> ratio = auxjad.simplified_time_signature_ratio((4, 8))
        >>> time_signature = abjad.TimeSignature(ratio)
        >>> format(time_signature)
        abjad.TimeSignature((2, 4))
        >>> ratio = auxjad.simplified_time_signature_ratio((1, 1))
        >>> time_signature = abjad.TimeSignature(ratio)
        >>> format(time_signature)
        abjad.TimeSignature((4, 4))

        If a ratio cannot be simplified at all, the function returns the
        original values.

        >>> ratio = auxjad.simplified_time_signature_ratio((7, 8))
        >>> time_signature = abjad.TimeSignature(ratio)
        >>> format(time_signature)
        abjad.TimeSignature((7, 8))

        The ``min_denominator`` can be set to values other than 4. If set to 2,
        the simplest representaion of the ratio (4, 8) becomes (1, 2).

        >>> ratio = auxjad.simplified_time_signature_ratio((4, 8),
        ...                                                min_denominator=2
        ...                                                )
        >>> time_signature = abjad.TimeSignature(ratio)
        >>> format(time_signature)
        abjad.TimeSignature((1, 2))
        >>> ratio = auxjad.simplified_time_signature_ratio((1, 1),
        ...                                                min_denominator=1
        ...                                                )
        >>> time_signature = abjad.TimeSignature(ratio)
        >>> format(time_signature)
        abjad.TimeSignature((1, 1))
    """
    if not isinstance(ratio, (tuple, abjad.Duration, abjad.TimeSignature)):
        raise TypeError("'ratio' must be 'tuple', 'abjad.Duration', or "
                        "'abjad.TimeSignature'")
    if not isinstance(min_denominator, int):
        raise TypeError("'min_denominator' must be 'int'")

    abjad_duration = False
    abjad_time_signature = False
    if isinstance(ratio, abjad.Duration):
        abjad_duration = True
        numerator, denominator = ratio.pair
    elif isinstance(ratio, abjad.TimeSignature):
        abjad_time_signature = True
        numerator, denominator = ratio.pair
    else:
        numerator, denominator = ratio
        if not isinstance(numerator, int):
            raise TypeError("'ratio's elements must be 'int'")
        if not isinstance(denominator, int):
            raise TypeError("'ratio's elements must be 'int'")

    while denominator < min_denominator:
        numerator *= 2
        denominator *= 2
    while True:
        if numerator % 2 == 0 and denominator % 2 == 0 and \
                denominator >= 2 * min_denominator:
            numerator /= 2
            denominator /= 2
        else:
            if abjad_duration:
                return abjad.Duration(int(numerator), int(denominator))
            elif abjad_time_signature:
                return abjad.TimeSignature((int(numerator), int(denominator)))
            else:
                return int(numerator), int(denominator)



import abjad
import auxjad


def test_remove_repeated_time_signatures_01():
    staff = abjad.Staff(r"c'4 d'8 | c'4 d'8")
    abjad.attach(abjad.TimeSignature((3, 8)), staff[0])
    abjad.attach(abjad.TimeSignature((3, 8)), staff[2])
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/8
            c'4
            d'8
            \time 3/8
            c'4
            d'8
        }
        ''')
    staff = auxjad.remove_repeated_time_signatures(staff)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/8
            c'4
            d'8
            c'4
            d'8
        }
        ''')


def test_remove_repeated_time_signatures_02():
    staff = abjad.Staff(r"c'4 d'8 | e'4. | c'4 d'8")
    abjad.attach(abjad.TimeSignature((3, 8)), staff[0])
    abjad.attach(abjad.TimeSignature((3, 8)), staff[3])
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/8
            c'4
            d'8
            e'4.
            \time 3/8
            c'4
            d'8
        }
        ''')
    staff = auxjad.remove_repeated_time_signatures(staff)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/8
            c'4
            d'8
            e'4.
            c'4
            d'8
        }
        ''')


def test_remove_repeated_time_signatures_03():
    staff = abjad.Staff([abjad.Note("c'2"),
                         abjad.Chord("<d' f'>2"),
                         abjad.Tuplet((2, 3), "g2 a2 b2"),
                         ])
    abjad.attach(abjad.TimeSignature((2, 2)), staff[0])
    abjad.attach(abjad.TimeSignature((2, 2)), staff[2][0])
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 2/2
            c'2
            <d' f'>2
            \times 2/3 {
                \time 2/2
                g2
                a2
                b2
            }
        }
        ''')
    staff = auxjad.remove_repeated_time_signatures(staff)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 2/2
            c'2
            <d' f'>2
            \times 2/3 {
                g2
                a2
                b2
            }
        }
        ''')



import abjad
import pytest
import auxjad


def test_LoopWindowByElements_01():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopWindowByElements(input_music, window_size=3)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }
        ''')
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 11/8
            d'2
            e'4
            f'2
            ~
            f'8
        }
        ''')
    notes = looper.get_current_window()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 11/8
            d'2
            e'4
            f'2
            ~
            f'8
        }
        ''')


def test_LoopWindowByElements_02():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopWindowByElements(input_music,
                                         window_size=3,
                                         step_size=1,
                                         max_steps=2,
                                         repetition_chance=0.25,
                                         forward_bias=0.2,
                                         head_position=0,
                                         omit_time_signature=False,
                                         )
    assert looper.window_size == 3
    assert looper.step_size == 1
    assert looper.max_steps == 2
    assert looper.repetition_chance == 0.25
    assert looper.forward_bias == 0.2
    assert looper.head_position == 0
    assert not looper.omit_time_signature
    looper.set_window_size(2)
    looper.set_step_size(2)
    looper.set_max_steps(3)
    looper.set_repetition_chance(0.1)
    looper.set_forward_bias(0.8)
    looper.set_head_position(2)
    looper.set_omit_time_signature(True)
    assert looper.window_size == 2
    assert looper.step_size == 2
    assert looper.max_steps == 3
    assert looper.repetition_chance == 0.1
    assert looper.forward_bias == 0.8
    assert looper.head_position == 2
    assert looper.omit_time_signature


def test_LoopWindowByElements_03():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopWindowByElements(input_music, window_size=3)
    assert looper.head_position == 0
    looper()
    assert looper.head_position == 0
    looper()
    assert looper.head_position == 1
    looper()
    assert looper.head_position == 2


def test_LoopWindowByElements_04():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopWindowByElements(input_music, window_size=3)
    assert len(looper) == 5


def test_LoopWindowByElements_05():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopWindowByElements(input_music, window_size=2)
    notes = looper.output_all()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 2/4
            c'4
            d'4
            \time 2/4
            d'4
            e'4
            \time 2/4
            e'4
            f'4
            \time 1/4
            f'4
        }
        ''')


def test_LoopWindowByElements_06():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopWindowByElements(input_music, window_size=3)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }
        ''')
    looper.set_window_size(4)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 19/8
            d'2
            e'4
            f'2
            ~
            f'8
            g'1
        }
        ''')


def test_LoopWindowByElements_07():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopWindowByElements(input_music,
                                         window_size=3,
                                         head_position=2,
                                         )
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 15/8
            e'4
            f'2
            ~
            f'8
            g'1
        }
        ''')


def test_LoopWindowByElements_08():
    input_music = abjad.Container(r"c'4 d'8 \times 2/3 {a4 g2}")
    looper = auxjad.LoopWindowByElements(input_music, window_size=2)
    notes = looper.output_all()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/8
            c'4
            d'8
            #(ly:expect-warning "strange time signature found")
            \time 7/24
            d'8
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                a4
            }
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                \time 2/4
                a4
            }
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                g2
            }
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                #(ly:expect-warning "strange time signature found")
                \time 2/6
                g2
            }
        }
        ''')


def test_LoopWindowByElements_09():
    input_music = abjad.Container(r"c'4 d'2")
    looper = auxjad.LoopWindowByElements(input_music, window_size=2)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/4
            c'4
            d'2
        }
        ''')
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 2/4
            d'2
        }
        ''')
    with pytest.raises(StopIteration):
        assert looper.__next__()


def test_LoopWindowByElements_10():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopWindowByElements(input_music,
                                         window_size=3,
                                         omit_time_signature=True,
                                         )
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            d'2
            e'4
        }
        ''')


def test_LoopWindowByElements_11():
    wrong_type_input = 'foo'
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    with pytest.raises(TypeError):
        assert auxjad.LoopWindowByElements(wrong_type_input, window_size=3)
        assert auxjad.LoopWindowByElements(input_music, window_size='foobar')
        assert auxjad.LoopWindowByElements(input_music,
                                           window_size=3,
                                           step_size='foobar',
                                           )
        assert auxjad.LoopWindowByElements(input_music,
                                           window_size=3,
                                           max_steps='foobar',
                                           )
        assert auxjad.LoopWindowByElements(input_music,
                                           window_size=3,
                                           repetition_chance='foobar',
                                           )
        assert auxjad.LoopWindowByElements(input_music,
                                           window_size=3,
                                           head_position='foobar',
                                           )
        assert auxjad.LoopWindowByElements(input_music,
                                           window_size=3,
                                           omit_time_signature='foobar',
                                           )
    with pytest.raises(ValueError):
        assert auxjad.LoopWindowByElements(input_music, window_size=100)
        assert auxjad.LoopWindowByElements(input_music,
                                           window_size=3,
                                           step_size=-1,
                                           )
        assert auxjad.LoopWindowByElements(input_music,
                                           window_size=3,
                                           step_size=100,
                                           )
        assert auxjad.LoopWindowByElements(input_music,
                                           window_size=3,
                                           max_steps=-1,
                                           )
        assert auxjad.LoopWindowByElements(input_music,
                                           window_size=3,
                                           repetition_chance=-0.3,
                                           )
        assert auxjad.LoopWindowByElements(input_music,
                                           window_size=3,
                                           repetition_chance=1.4,
                                           )
        assert auxjad.LoopWindowByElements(input_music,
                                           window_size=3,
                                           head_position=-1,
                                           )
        assert auxjad.LoopWindowByElements(input_music,
                                           window_size=3,
                                           head_position=100,
                                           )


def test_LoopWindowByElements_12():
    input_music = abjad.Container(r"c'4 d'2 r8 d'4 <e' g'>8 r4 f'2. <e' g'>16")
    looper = auxjad.LoopWindowByElements(input_music,
                                         window_size=4,
                                         )
    music = looper.output_all(tie_identical_pitches=True)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 9/8
            c'4
            d'2
            r8
            d'4
            ~
            \time 4/4
            d'2
            r8
            d'4
            <e' g'>8
            \time 3/4
            r8
            d'4
            <e' g'>8
            r4
            \time 11/8
            d'4
            <e' g'>8
            r4
            f'2.
            \time 19/16
            <e' g'>8
            r4
            f'2.
            <e' g'>16
            \time 17/16
            r4
            f'2.
            <e' g'>16
            \time 13/16
            f'2.
            <e' g'>16
            ~
            \time 1/16
            <e' g'>16
        }
        ''')


def test_LoopWindowByElements_13():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopWindowByElements(input_music, window_size=2)
    notes = looper.output_n(2)
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 2/4
            c'4
            d'4
            \time 2/4
            d'4
            e'4
        }
        ''')


def test_LoopWindowByElements_14():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopWindowByElements(input_music, window_size=2)
    notes = looper.output_n(2, tie_identical_pitches=True)
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 2/4
            c'4
            d'4
            ~
            \time 2/4
            d'4
            e'4
        }
        ''')


def test_LoopWindowByElements_15():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopWindowByElements(input_music, window_size=2)
    with pytest.raises(RuntimeError):
        looper.output_n(100)


def test_LoopWindowByElements_16():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopWindowByElements(input_music,
                                         window_size=2,
                                         head_position=2,
                                         forward_bias=0.0
                                         )
    notes = looper.output_all()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 2/4
            e'4
            f'4
            \time 2/4
            d'4
            e'4
            \time 2/4
            c'4
            d'4
        }
        ''')


def test_LoopWindowByElements_16():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopWindowByElements(input_music,
                                     window_size=2,
                                     head_position=0,
                                     forward_bias=0.0
                                     )
    notes = looper.output_all()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
    r'''
    \new Staff
    {
        \time 2/4
        c'4
        d'4
    }
    ''')



import copy
import abjad
from ._LoopWindowGeneric import _LoopWindowGeneric


class LoopWindow(_LoopWindowGeneric):
    r"""Using a looping window, this slices an input ``abjad.Container`` and
    output them as containers.

    ..  container:: example

        Usage is similar to other factory classes. It takes a container (or
        child class equivalent) as argument. Each call of the object, in this
        case ``looper()``, will move the window forwards and output the sliced
        window. If no window_size nor step_size are entered as arguments, they
        are set to the following default values, respectively: (4, 4), i.e. a
        window of the size of a 4/4 bar, and (1, 16), i.e. a step of the length
        of a sixteenth-note.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopWindow(input_music)
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }

        The method ``get_current_window()`` will output the current window
        without moving the head forwards.

        >>> notes = looper.get_current_window()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }

    ..  container:: example

        The optional arguments ``window_size`` and ``step_size`` can be used to
        set different window and step sizes. ``window_size`` can take a tuple
        or an ``abjad.Meter`` as input, while ``step_size`` takes a tuple or an
        ``abjad.Duration``.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopWindow(input_music,
        ...                            window_size=(3, 4),
        ...                            step_size=(1, 4),
        ...                            )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'4
            d'2
        }
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            d'2
            e'4
        }

    ..  container:: example

        The instances of ``LoopWindow`` can also be used as an iterator, which
        can then be used in a for loop to exhaust all windows. Notice how it
        appends rests at the end of the container, until it is totally
        exhausted.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4")
        >>> looper = auxjad.LoopWindow(input_music,
        ...                            window_size=(3, 4),
        ...                            step_size=(1, 8),
        ...                            )
        >>> for window in looper:
        ...     staff = abjad.Staff(window)
        ...     abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'4
            d'2
        }
        \new Staff
        {
            c'8
            d'8
            ~
            d'4.
            e'8
        }
        \new Staff
        {
            d'2
            e'4
        }
        \new Staff
        {
            d'4.
            e'8
            ~
            e'8
            r8
        }
        \new Staff
        {
            d'4
            e'4
            r4
        }
        \new Staff
        {
            d'8
            e'8
            ~
            e'8
            r4.
        }
        \new Staff
        {
            e'4
            r2
        }
        \new Staff
        {
            e'8
            r8
            r2
        }

    ..  container:: example

        This class can take many optional keyword arguments during its
        creation, besides ``window_size`` and ``step_size``. ``max_steps`` sets
        the maximum number of steps that the window can advance when the object
        is called, ranging between 1 and the input value (default is also 1).
        ``repetition_chance`` sets the chance of a window result repeating
        itself (that is, the window not moving forwards when called). It should
        range from 0.0 to 1.0 (default 0.0, i.e. no repetition).
        ``forward_bias`` sets the chance of the window moving forward instead
        of backwards. It should range from 0.0 to 1.0 (default 1.0, which means
        the window can only move forwards. A value of 0.5 gives 50% chance of
        moving forwards while a value of 0.0 will move the window only
        backwards). Finally, ``head_position`` can be used to offset the
        starting position of the looping window. It must be a tuple or an
        ``abjad.Duration``, and its default value is 0.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopWindow(input_music,
        ...                            window_size=(3, 4),
        ...                            step_size=(5, 8),
        ...                            max_steps=2,
        ...                            repetition_chance=0.25,
        ...                            forward_bias=0.2,
        ...                            head_position=(2, 8),
        ...                            omit_time_signature=False,
        ...                            )
        >>> looper.window_size
        3/4
        >>> looper.step_size
        5/8
        >>> looper.repetition_chance
        0.25
        >>> looper.forward_bias
        0.2
        >>> looper.max_steps
        2
        >>> looper.head_position
        1/4
        >>> looper.omit_time_signature
        False

        Use the ``set_`` methods below to change these values after
        initialisation.

        >>> looper.set_window_size((5, 4))
        >>> looper.set_step_size((1, 4))
        >>> looper.set_max_steps(3)
        >>> looper.set_repetition_chance(0.1)
        >>> looper.set_forward_bias(0.8)
        >>> looper.set_head_position(0)
        >>> looper.set_omit_time_signature(True)
        >>> looper.window_size
        5/4
        >>> looper.step_size
        1/4
        >>> looper.max_steps
        3
        >>> looper.repetition_chance
        0.1
        >>> looper.forward_bias
        0.8
        >>> looper.head_position
        0
        >>> looper.omit_time_signature
        True

    ..  container:: example

        To run through the whole process and output it as a single container,
        from the initial head position until the process outputs the single
        last element, use the method ``output_all()``.

        >>> input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LoopWindow(input_music,
        ...                            window_size=(3, 4),
        ...                            step_size=(1, 4),
        ...                            )
        >>> music = looper.output_all()
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'4
            d'4
            e'4
            d'4
            e'4
            f'4
            e'4
            f'4
            r4
            f'4
            r2
        }

    ..  container:: example

        When using ``output_all()``, set the keyword argument
        ``tie_identical_pitches`` to ``True`` in order to tie identical notes
        or chords at the end and beginning of consecutive windows.

        >>> input_music = abjad.Container(r"c'4 <e' f' g'>2 r4 f'2.")
        >>> looper = auxjad.LoopWindow(input_music,
        ...                            window_size=(3, 4),
        ...                            step_size=(1, 4),
        ...                            )
        >>> music = looper.output_all(tie_identical_pitches=True)
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'4
            <e' f' g'>2
            ~
            <e' f' g'>2
            r4
            <e' f' g'>4
            r4
            f'4
            r4
            f'2
            ~
            f'2.
            ~
            f'2
            r4
            f'4
            r2
        }

    ..  container:: example

        To run through just part of the process and output it as a single
        container, starting from the initial head position, use the method
        ``output_n()`` and pass the number of iterations as argument. Similarly
        to ``output_all()``, the keyword argument ``tie_identical_pitches`` is
        available for tying pitches.

        >>> input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LoopWindow(input_music,
        ...                            window_size=(3, 4),
        ...                            step_size=(1, 4),
        ...                            )
        >>> music = looper.output_n(2)
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'4
            d'4
            e'4
            d'4
            e'4
            f'4
        }

    .. container:: example

        To change the size of the looping window after instantiation, use the
        method ``set_window_size()``. In the example below, the initial window
        is of size (4, 4), but changes to (3, 8) after three calls. Notice how
        the very first call attaches a time signature equivalent to the window
        size to the output window; subsequent calls will not have time
        signatures unless the size of the looping window changes.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopWindow(input_music)
        >>> for _ in range(3):
        ...     notes = looper()
        ...     staff = abjad.Staff(notes)
        ...     abjad.f(staff)
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }
        \new Staff
        {
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }
        \new Staff
        {
            c'8
            d'8
            ~
            d'4.
            e'8
            ~
            e'8
            f'8
        }
        >>> looper.set_window_size((3, 8))
        >>> for _ in range(3):
        ...     notes = looper()
        ...     staff = abjad.Staff(notes)
        ...     abjad.f(staff)
        \new Staff
        {
            \time 3/8
            c'16
            d'16
            ~
            d'4
        }
        \new Staff
        {
            d'4.
        }
        \new Staff
        {
            d'4.
        }

        To disable time signatures altogether, initialise ``LoopWindow`` with
        the keyword argument ``omit_time_signature`` set to ``True`` (default
        is ``False``), or use the ``set_omit_time_signature()`` method after
        initialisation.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopWindow(input_music, omit_time_signature=True)
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            d'2
            e'4
        }

    ..  container:: example

        This class can handle tuplets, but this functionality should be
        considered experimental.

        >>> input_music = abjad.Container(r"\times 2/3 {c'8 d'8 e'} d'2.")
        >>> looper = auxjad.LoopWindow(input_music,
        ...                            window_size=(3, 4),
        ...                            step_size=(1, 16))
        >>> staff = abjad.Staff()
        >>> for _ in range(3):
        ...     window = looper()
        ...     staff.append(window)
        >>> abjad.f(staff)
        \new Staff
        {
            \times 2/3 {
                \time 3/4
                c'8
                d'8
                e'8
            }
            d'2
            \times 2/3 {
                c'32
                d'16
                ~
                d'16
                e'8
            }
            d'16
            ~
            d'2
            \times 2/3 {
                d'16
                e'8
            }
            d'8
            ~
            d'2
        }
    """

    def __init__(self,
                 container: abjad.Container,
                 *,
                 window_size: (tuple, abjad.Meter) = (4, 4),
                 step_size: (int, float, tuple, str, abjad.Duration) = (1, 16),
                 max_steps: int = 1,
                 repetition_chance: float = 0.0,
                 forward_bias: float = 1.0,
                 head_position: (int, float, tuple, str, abjad.Duration) = 0,
                 omit_time_signature: bool = False,
                 ):
        if not isinstance(container, abjad.Container):
            raise TypeError("'container' must be 'abjad.Container' or child "
                            "class")
        self._container = copy.deepcopy(container)
        self._container_length = abjad.inspect(container[:]).duration()
        self._new_time_signature = True
        self.set_omit_time_signature(omit_time_signature)
        super().__init__(head_position,
                         window_size,
                         step_size,
                         max_steps,
                         repetition_chance,
                         forward_bias,
                         )

    def set_head_position(self,
                          head_position: tuple,
                          ):
        r"""Custom ``set_head_position()`` method since parent's method uses
        integers as input, intead of tuples or ``abjad.Duration``.
        """
        if not isinstance(head_position,
                          (int, float, tuple, str, abjad.Duration),
                          ):
            raise TypeError("'head_position' must be a number or "
                            "'abjad.Duration'")
        if abjad.Duration(head_position) >= self._container_length:
            raise ValueError("'head_position' must be smaller than the "
                             "length of 'container'")
        self.head_position = abjad.Duration(head_position)

    def set_window_size(self,
                        window_size: (tuple, abjad.Meter),
                        ):
        r"""Custom set_window_size() method since parent's method uses
        integers as input, intead of tuples or abjad.Duration.
        """
        if not isinstance(window_size,
                          (int, float, tuple, str, abjad.Meter),
                          ):
            raise TypeError("'window_size' must be 'tuple' or 'abjad.Meter'")
        if abjad.Meter(window_size).duration > self._container_length \
                              - self.head_position:
            raise ValueError("'window_size' must be smaller than or equal "
                             "to the length of 'container'")
        if self._first_window or self.window_size.duration \
                != abjad.Meter(window_size).duration:
            self.window_size = abjad.Meter(window_size)
            self._new_time_signature = True

    def set_step_size(self,
                      step_size: tuple,
                      ):
        r"""Custom ``set_step_size()`` method since parent's method uses
        integers as input, intead of tuples or ``abjad.Duration``.
        """
        if not isinstance(step_size,
                          (int, float, tuple, str, abjad.Duration),
                          ):
            raise TypeError("'step_size' must be a 'tuple' or "
                            "'abjad.Duration'")
        self.step_size = abjad.Duration(step_size)

    def set_omit_time_signature(self,
                                omit_time_signature: bool,
                                ):
        if not isinstance(omit_time_signature, bool):
            raise TypeError("'omit_time_signature' must be 'bool'")
        self.omit_time_signature = omit_time_signature

    def _done(self) -> bool:
        r"""Custom ``_done`` method since parent's method uses the ``__len__``
        method, which cannot be used for non-integer values such as
        ``abjad.Duration``.
        """
        return self.head_position >= self._container_length or \
            self.head_position < 0

    def _slice_container(self) -> abjad.Selection:
        head = self.head_position
        window_size = self.window_size
        dummy_container = copy.deepcopy(self._container)
        # splitting leaves at both slicing points
        if head > abjad.Duration(0):
            abjad.mutate(dummy_container[:]).split([head,
                                                    window_size.duration,
                                                    ])
        else:
            abjad.mutate(dummy_container[:]).split([window_size.duration])
        # finding start and end indeces for the window
        for start in range(len(dummy_container)):
            if abjad.inspect(dummy_container[:start + 1]).duration() > head:
                abjad.detach(abjad.Tie(), dummy_container[start - 1])
                break
        for end in range(start + 1, len(dummy_container)):
            if abjad.inspect(dummy_container[start : end]).duration() == \
                    window_size.duration:
                abjad.detach(abjad.Tie(), dummy_container[end - 1])
                break
        else:
            end = len(dummy_container)
        # appending rests if necessary
        container_dur = abjad.inspect(dummy_container[start : end]).duration()
        if container_dur < window_size.duration:
            missing_dur = window_size.duration - container_dur
            rests = abjad.LeafMaker()(None, missing_dur)
            dummy_container.extend(rests)
            end += len(rests)
        # transforming abjad.Selection -> abjad.Container for rewrite_meter
        dummy_container = abjad.Container(
            abjad.mutate(dummy_container[start : end]).copy()
        )
        abjad.mutate(dummy_container[:]).rewrite_meter(window_size)
        if self._new_time_signature and not self.omit_time_signature:
            abjad.attach(abjad.TimeSignature(window_size),
                         abjad.select(dummy_container).leaves()[0],
                         )
            self._new_time_signature = False
        self._current_window = dummy_container[:]
        dummy_container[:] = []



import abjad
import pytest
import auxjad


def test_LoopWindow_01():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopWindow(input_music)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }
        ''')
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }
        ''')
    notes = looper.get_current_window()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }
        ''')


def test_LoopWindow_02():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopWindow(input_music,
                               window_size=(3, 4),
                               step_size=(1, 4),
                               )
    notes = looper()
    staff = abjad.Staff(notes)
    abjad.f(staff)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/4
            c'4
            d'2
        }
        ''')
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            d'2
            e'4
        }
        ''')


def test_LoopWindow_03():
    input_music = abjad.Container(r"c'4 d'2 e'4")
    looper = auxjad.LoopWindow(input_music,
                               window_size=(3, 4),
                               step_size=(1, 8),
                               )
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/4
            c'4
            d'2
        }
        ''')
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'8
            d'8
            ~
            d'4.
            e'8
        }
        ''')
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            d'2
            e'4
        }
        ''')
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            d'4.
            e'8
            ~
            e'8
            r8
        }
        ''')
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            d'4
            e'4
            r4
        }
        ''')
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            d'8
            e'8
            ~
            e'8
            r4.
        }
        ''')
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            e'4
            r2
        }
        ''')
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            e'8
            r8
            r2
        }
        ''')
    with pytest.raises(StopIteration):
        assert looper.__next__()


def test_LoopWindow_04():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopWindow(input_music,
                               window_size=(3, 4),
                               step_size=(5, 8),
                               max_steps=2,
                               repetition_chance=0.25,
                               forward_bias=0.2,
                               head_position=(2, 8),
                               omit_time_signature=False,
                               )
    assert looper.window_size == abjad.Meter((3, 4))
    assert looper.step_size == abjad.Duration((5, 8))
    assert looper.max_steps == 2
    assert looper.repetition_chance == 0.25
    assert looper.forward_bias == 0.2
    assert looper.head_position == abjad.Duration((1, 4))
    assert not looper.omit_time_signature
    looper.set_window_size((5, 4))
    looper.set_step_size((1, 4))
    looper.set_max_steps(3)
    looper.set_repetition_chance(0.1)
    looper.set_forward_bias(0.8)
    looper.set_head_position(0)
    looper.set_omit_time_signature(True)
    assert looper.window_size == abjad.Meter((5, 4))
    assert looper.step_size == abjad.Duration((1, 4))
    assert looper.max_steps == 3
    assert looper.repetition_chance == 0.1
    assert looper.forward_bias == 0.8
    assert looper.head_position == abjad.Duration(0)
    assert looper.omit_time_signature


def test_LoopWindow_05():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopWindow(input_music,
                               window_size=(3, 4),
                               step_size=(1, 4),
                               )
    music = looper.output_all()
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/4
            c'4
            d'4
            e'4
            d'4
            e'4
            f'4
            e'4
            f'4
            r4
            f'4
            r2
        }
        ''')


def test_LoopWindow_06():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopWindow(input_music)
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }
        ''')
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'8.
            d'16
            ~
            d'4..
            e'16
            ~
            e'8.
            f'16
        }
        ''')
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'8
            d'8
            ~
            d'4.
            e'8
            ~
            e'8
            f'8
        }
        ''')
    looper.set_window_size((3, 8))
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/8
            c'16
            d'16
            ~
            d'4
        }
        ''')
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            d'4.
        }
        ''')
    notes = looper.__next__()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            d'4.
        }
        ''')


def test_LoopWindow_07():
    input_music = abjad.Container(r"\times 2/3 {c'8 d'8 e'} d'2.")
    looper = auxjad.LoopWindow(input_music,
                               window_size=(3, 4),
                               step_size=(1, 16))
    staff = abjad.Staff()
    for _ in range(3):
        window = looper()
        staff.append(window)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \times 2/3 {
                \time 3/4
                c'8
                d'8
                e'8
            }
            d'2
            \times 2/3 {
                c'32
                d'16
                ~
                d'16
                e'8
            }
            d'16
            ~
            d'2
            \times 2/3 {
                d'16
                e'8
            }
            d'8
            ~
            d'2
        }
        ''')


def test_LoopWindow_08():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    looper = auxjad.LoopWindow(input_music, omit_time_signature=True)
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            d'2
            e'4
        }
        ''')


def test_LoopWindow_09():
    wrong_type_input = 'foobar'
    input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
    with pytest.raises(TypeError):
        assert auxjad.LoopWindow(wrong_type_input)
        assert auxjad.LoopWindow(input_music, window_size=17j)
        assert auxjad.LoopWindow(input_music, max_steps='foobar')
        assert auxjad.LoopWindow(input_music, repetition_chance='foobar')
        assert auxjad.LoopWindow(input_music, head_position=62.3j)
        assert auxjad.LoopWindow(input_music, omit_time_signature='foobar')
    with pytest.raises(ValueError):
        assert auxjad.LoopWindow(input_music, window_size=(100, 1))
        assert auxjad.LoopWindow(input_music, max_steps=-1)
        assert auxjad.LoopWindow(input_music, repetition_chance=-0.3)
        assert auxjad.LoopWindow(input_music, repetition_chance=1.4)
        assert auxjad.LoopWindow(input_music, head_position=(100, 1))


def test_LoopWindow_10():
    input_music = abjad.Container(r"c'4 e'2 d'2 f'4")
    looper = auxjad.LoopWindow(input_music,
                               window_size=(3, 4),
                               step_size=(1, 4),
                               )
    music = looper.output_all(tie_identical_pitches=True)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/4
            c'4
            e'2
            ~
            e'2
            d'4
            e'4
            d'2
            ~
            d'2
            f'4
            d'4
            f'4
            r4
            f'4
            r2
        }
        ''')


def test_LoopWindow_11():
    input_music = abjad.Container(r"c'4 <e' f' g'>2 r4 f'2.")
    looper = auxjad.LoopWindow(input_music,
                               window_size=(3, 4),
                               step_size=(1, 4),
                               )
    music = looper.output_all(tie_identical_pitches=True)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/4
            c'4
            <e' f' g'>2
            ~
            <e' f' g'>2
            r4
            <e' f' g'>4
            r4
            f'4
            r4
            f'2
            ~
            f'2.
            ~
            f'2
            r4
            f'4
            r2
        }
        ''')


def test_LoopWindow_12():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopWindow(input_music,
                               window_size=(3, 4),
                               step_size=(1, 4),
                               )
    music = looper.output_n(2)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/4
            c'4
            d'4
            e'4
            d'4
            e'4
            f'4
        }
        ''')


def test_LoopWindow_13():
    input_music = abjad.Container(r"c'4 d'2 e'4 f'4")
    looper = auxjad.LoopWindow(input_music,
                               window_size=(3, 4),
                               step_size=(1, 4),
                               )
    music = looper.output_n(2, tie_identical_pitches=True)
    staff = abjad.Staff(music)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/4
            c'4
            d'2
            ~
            d'2
            e'4
        }
        ''')


def test_LoopWindow_14():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
    looper = auxjad.LoopWindow(input_music,
                               window_size=(3, 4),
                               step_size=(1, 4),
                               )
    with pytest.raises(RuntimeError):
        looper.output_n(100)


def test_LoopWindow_15():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4 g'4 a'4")
    looper = auxjad.LoopWindow(input_music,
                               window_size=(3, 4),
                               step_size=(1, 4),
                               head_position=(3, 4),
                               forward_bias = 0.0
                               )
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            \time 3/4
            f'4
            g'4
            a'4
        }
        ''')
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            e'4
            f'4
            g'4
        }
        ''')
    notes = looper()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            d'4
            e'4
            f'4
        }
        ''')


def test_LoopWindow_16():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4 g'4 a'4")
    looper = auxjad.LoopWindow(input_music,
                           window_size=(3, 4),
                           step_size=(1, 4),
                           head_position=(3, 4),
                           forward_bias=0.0,
                           )
    notes = looper.output_all()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
    r'''
    \new Staff
    {
        \time 3/4
        f'4
        g'4
        a'4
        e'4
        f'4
        g'4
        d'4
        e'4
        f'4
        c'4
        d'4
        e'4
    }
    ''')


def test_LoopWindow_17():
    input_music = abjad.Container(r"c'4 d'4 e'4 f'4 g'4 a'4")
    looper = auxjad.LoopWindow(input_music,
                           window_size=(3, 4),
                           step_size=(1, 4),
                           forward_bias=0.0,
                           )
    notes = looper.output_all()
    staff = abjad.Staff(notes)
    assert format(staff) == abjad.String.normalize(
    r'''
    \new Staff
    {
        \time 3/4
        c'4
        d'4
        e'4
    }
    ''')



import random
import abjad
import auxjad


def test_CartographyContainer_01():
    random.seed(41298)
    container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
    assert container.contents == [0, 1, 2, 3, 4]
    assert container.weights == [1.0, 0.75, 0.5625, 0.421875, 0.31640625]
    assert len(container) == 5
    result = ''
    for _ in range(30):
        result += str(container())
    assert result == '203001402200011111101400310140'


def test_CartographyContainer_02():
    container = auxjad.CartographyContainer([0, 1, 2, 3, 4], decay_rate=0.5)
    assert container.weights == [1.0, 0.5, 0.25, 0.125, 0.0625]


def test_CartographyContainer_03():
    container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
    container.append(5)
    assert container.contents == [1, 2, 3, 4, 5]
    container.append(42)
    assert container.contents == [2, 3, 4, 5, 42]


def test_CartographyContainer_04():
    container = auxjad.CartographyContainer([10, 7, 14, 31, 98])
    container.append_keeping_n(100, 2)
    assert container.contents == [10, 7, 31, 98, 100]


def test_CartographyContainer_05():
    container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
    container.prepend(-1)
    assert container.contents == [-1, 0, 1, 2, 3]
    container.prepend(71)
    assert container.contents == [71, -1, 0, 1, 2]


def test_CartographyContainer_06():
    container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
    container.rotate()
    assert container.contents == [1, 2, 3, 4, 0]
    container.rotate(anticlockwise=True)
    assert container.contents == [0, 1, 2, 3, 4]
    container.rotate(anticlockwise=True)
    assert container.contents == [4, 0, 1, 2, 3]


def test_CartographyContainer_07():
    random.seed(15424)
    container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
    container.randomise()
    assert container.contents == [1, 4, 3, 0, 2]


def test_CartographyContainer_08():
    container = auxjad.CartographyContainer([0, 1, 2, 3, 4], decay_rate=0.5)
    assert len(container) == 5
    assert container.weights == [1.0, 0.5, 0.25, 0.125, 0.0625]
    container.set_container([10, 7, 14, 31, 98, 47, 32])
    assert container.contents == [10, 7, 14, 31, 98, 47, 32]
    assert len(container) == 7
    assert container.weights == [1.0,
                                 0.5,
                                 0.25,
                                 0.125,
                                 0.0625,
                                 0.03125,
                                 0.015625,
                                 ]


def test_CartographyContainer_09():
    container = auxjad.CartographyContainer([10, 7, 14, 31, 98])
    container.replace_element(100, 2)
    assert container.contents == [10, 7, 100, 31, 98]


def test_CartographyContainer_10():
    random.seed(83552)
    container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
    container.set_decay_rate(0.2)
    assert container.weights == [1.0,
                                 0.2,
                                 0.04000000000000001,
                                 0.008000000000000002,
                                 0.0016000000000000003,
                                 ]
    result = ''
    for _ in range(30):
        result += str(container())
    assert result == '000001002100000201001030000100'


def test_CartographyContainer_11():
    random.seed(19844)
    container = auxjad.CartographyContainer([10, 7, 14, 31, 98])
    assert container() == 31
    n = container.previous_index
    assert n == 3
    assert container.get_element(n) == 31


def test_CartographyContainer_12():
    random.seed(98743)
    container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
    result = ''
    for _ in range(30):
        result += str(container())
    assert result == '210431340000344203001220034203'
    result = ''
    for _ in range(30):
        result += str(container(no_repeat=True))
    assert result == '210421021020304024230120241202'


def test_CartographyContainer_13():
    container = auxjad.CartographyContainer([10, 7, 14, 31, 98])
    assert container.contents[2] == 14
    assert container.contents[1:4] == [7, 14, 31]


def test_CartographyContainer_14():
    container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
    container.mirror(0)
    assert container.contents == [4, 1, 2, 3, 0]
    container.mirror(0)
    assert container.contents == [0, 1, 2, 3, 4]
    container.mirror(3)
    assert container.contents == [0, 3, 2, 1, 4]
    container.mirror(2)
    assert container.contents == [0, 3, 2, 1, 4]


def test_CartographyContainer_15():
    random.seed(90129)
    container = auxjad.CartographyContainer([0, 1, 2, 3, 4])
    container.mirror_random()
    assert container.contents == [4, 1, 2, 3, 0]
    container.mirror_random()
    assert container.contents == [4, 3, 2, 1, 0]
    container.mirror_random()
    assert container.contents == [4, 1, 2, 3, 0]



import abjad
import auxjad


def test_remove_repeated_dynamics_01():
    staff = abjad.Staff(r"c'4\pp d'8\pp | c'4\f d'8\f")
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \pp
            d'8
            \pp
            c'4
            \f
            d'8
            \f
        }
        ''')
    auxjad.remove_repeated_dynamics(staff)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \pp
            d'8
            c'4
            \f
            d'8
        }
        ''')


def test_remove_repeated_dynamics_02():
    staff = abjad.Staff(r"c'4\p d'8 | e'4.\p | c'4\p d'8\f")
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \p
            d'8
            e'4.
            \p
            c'4
            \p
            d'8
            \f
        }
        ''')
    auxjad.remove_repeated_dynamics(staff)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \p
            d'8
            e'4.
            c'4
            d'8
            \f
        }
        ''')


def test_remove_repeated_dynamics_03():
    staff = abjad.Staff([abjad.Note("c'2"),
                         abjad.Chord("<d' f'>2"),
                         abjad.Tuplet((2, 3), "g2 a2 b2"),
                         ])
    abjad.attach(abjad.Dynamic('ppp'), staff[0])
    abjad.attach(abjad.Dynamic('ppp'), staff[1])
    abjad.attach(abjad.Dynamic('ppp'), staff[2][0])
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'2
            \ppp
            <d' f'>2
            \ppp
            \times 2/3 {
                g2
                \ppp
                a2
                b2
            }
        }
        ''')
    auxjad.remove_repeated_dynamics(staff)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'2
            \ppp
            <d' f'>2
            \times 2/3 {
                g2
                a2
                b2
            }
        }
        ''')


def test_remove_repeated_dynamics_04():
    staff = abjad.Staff(r"c'4\pp\< d'8\f\> | c'4\f d'8\f")
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \pp
            \<
            d'8
            \f
            \>
            c'4
            \f
            d'8
            \f
        }
        ''')
    auxjad.remove_repeated_dynamics(staff)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \pp
            \<
            d'8
            \f
            \>
            c'4
            \f
            d'8
        }
        ''')


def test_remove_repeated_dynamics_05():
    staff = abjad.Staff(r"c'4\pp\< d'8\f\> | c'4\f d'8\f")
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \pp
            \<
            d'8
            \f
            \>
            c'4
            \f
            d'8
            \f
        }
        ''')
    auxjad.remove_repeated_dynamics(staff, ignore_hairpins=True)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \pp
            \<
            d'8
            \f
            \>
            c'4
            d'8
        }
        ''')


def test_remove_repeated_dynamics_06():
    staff = abjad.Staff(r"c'4\pp r2. | c'4\pp")
    auxjad.remove_repeated_dynamics(staff)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \pp
            r2.
            c'4
        }
        ''')
    staff = abjad.Staff(r"c'4\pp r2. | c'4\pp")
    auxjad.remove_repeated_dynamics(staff, reset_after_rests=True)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \pp
            r2.
            c'4
            \pp
        }
        ''')


def test_remove_repeated_dynamics_07():
    staff = abjad.Staff(r"c'4\pp r2. | c'4\pp")
    auxjad.remove_repeated_dynamics(staff, reset_after_rests=(4, 4))
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \pp
            r2.
            c'4
        }
        ''')
    staff = abjad.Staff(r"c'4\pp r2. | c'4\pp")
    auxjad.remove_repeated_dynamics(staff, reset_after_rests=2/4)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \pp
            r2.
            c'4
            \pp
        }
        ''')


def test_remove_repeated_dynamics_08():
    staff = abjad.Staff(r"c'4\pp r2. | R1 | c'4\pp")
    auxjad.remove_repeated_dynamics(staff,
                                    reset_after_rests=abjad.Duration(4, 4),
                                    )
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \pp
            r2.
            R1
            c'4
            \pp
        }
        ''')
    staff = abjad.Staff(r"c'4\pp r2. | R1 | c'4\pp")
    auxjad.remove_repeated_dynamics(staff, reset_after_rests=2)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            c'4
            \pp
            r2.
            R1
            c'4
        }
        ''')



from .are_containers_equal import are_containers_equal
from .are_leaves_tieable import are_leaves_tieable
from .CartographyContainer import CartographyContainer
from .close_container import close_container
from .fill_with_rests import fill_with_rests
from .is_container_full import is_container_full
from .LeafDynMaker import LeafDynMaker
from .LoopWindow import LoopWindow
from .LoopWindowByElements import LoopWindowByElements
from .LoopWindowByList import LoopWindowByList
from .remove_repeated_dynamics import remove_repeated_dynamics
from .remove_repeated_time_signatures import remove_repeated_time_signatures
from .simplified_time_signature_ratio import simplified_time_signature_ratio
from .sync_containers import sync_containers
from .TenneysContainer import TenneysContainer
from .underfull_duration import underfull_duration



import abjad
from .underfull_duration import underfull_duration


def is_container_full(container: abjad.Container) -> bool:
    r"""Checks if an ``abjad.Container`` is full. Based on
    ``auxjad.underfull_duration``.

    ..  container:: example

        Returns ``True`` if the last bar of any container (or child class) is
        full, otherwise returns ``False``. If no time signature is encountered,
        it uses LilyPond's convention and considers the container as in 4/4.

        >>> container1 = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Container(r"c'4 d'4 e'4")
        >>> container3 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4")
        >>> container4 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4 d'4 e'4 f'4")
        >>> auxjad.is_container_full(container1)
        True
        >>> auxjad.is_container_full(container2)
        False
        >>> auxjad.is_container_full(container3)
        False
        >>> auxjad.is_container_full(container4)
        True

    ..  container:: example

        Handles any time signatures as well as changes of time signature.

        >>> container1 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Container(r"\time 3/4 a2. \time 2/4 r2")
        >>> container3 = abjad.Container(r"\time 5/4 g1 ~ g4 \time 4/4 af'2")
        >>> container4 = abjad.Container(r"\time 6/8 c'2 ~ c'8")
        >>> auxjad.is_container_full(container1)
        True
        >>> auxjad.is_container_full(container2)
        True
        >>> auxjad.is_container_full(container3)
        False
        >>> auxjad.is_container_full(container4)
        False

    ..  container:: example

        Correctly handles partial time signatures.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> time_signature = abjad.TimeSignature((3, 4), partial=(1, 4))
        >>> abjad.attach(time_signature, container[0])
        >>> auxjad.is_container_full(container)
        True

    ..  container:: example

        It also handles multi-measure rests.

        >>> container1 = abjad.Container(r"R1")
        >>> container2 = abjad.Container(r"\time 3/4 R1*3/4 \time 2/4 r2")
        >>> container3 = abjad.Container(r"\time 5/4 R1*5/4 \time 4/4 g''4")
        >>> container4 = abjad.Container(r"\time 6/8 R1*1/2")
        >>> auxjad.is_container_full(container1)
        True
        >>> auxjad.is_container_full(container2)
        True
        >>> auxjad.is_container_full(container3)
        False
        >>> auxjad.is_container_full(container4)
        False

    ..  container:: example

        If a container is malformed, i.e. it has an underfilled bar before a
        time signature change, the function raises a ``ValueError`` exception.

        >>> container = abjad.Container(r"\time 5/4 g''1 \time 4/4 f'1")
        >>> auxjad.is_container_full(container)
        ValueError: 'container' is malformed, with an underfull bar preceeding
        a time signature change
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("'container' must be 'abjad.Container' or child class")
    return underfull_duration(container) == abjad.Duration(0)



"""auxjad
======

Auxiliary functions and classes for Abjad 3.1. All classes and functions have a
`__doc__` attribute with usage instructions.

Documentation is available at https://gilbertohasnofb.github.io/auxjad-docs/.
A pdf version of the documentation is also available in the `docs` directory.

Bugs can be reported to https://github.com/gilbertohasnofb/auxjad/issues.

This library is published under the MIT License.
"""

from .tools import *

__author__ = "Gilberto Agostinho <gilbertohasnofb@gmail.com>"
__version__ = "0.4.0"



import abjad


def remove_repeated_dynamics(container: abjad.Container,
                             *,
                             ignore_hairpins: bool = False,
                             reset_after_rests: bool = False,
                             ):
    r"""A function which removes all consecutive repeated dynamics. It removes
    consecutive effective dynamics, even if separated by any number of
    notes without one. It resets its memory of what was the previous dynamic
    every time it finds a hairpin, since notation such as ``"c'4\f\> c'4\f\>"``
    is quite common; this behaviour can be toggled off using the
    ``ignore_hairpins`` keyword argument. By default, it remembers the previous
    dynamic even with notes separated by rests; this can be toggled off using
    ``reset_after_rests=True``. To set a maximum length of silence after which
    dynamics are restated, set ``reset_after_rests`` to a duration using
    ``abjad.Duration()`` or any other duration format accepted by Abjad.

    ..  container:: example

        When two consecutive leaves have identical dynamics, the second
        one is removed:

        >>> staff = abjad.Staff(r"c'4\pp d'8\pp | c'4\f d'8\f")
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \pp
            d'8
            \pp
            c'4
            \f
            d'8
            \f
        }
        >>> auxjad.remove_repeated_dynamics(staff)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \pp
            d'8
            c'4
            \f
            d'8
        }

    ..  container:: example

        The function also removes dynamics that are separated by an arbitrary
        number of leaves without dynamics:

        >>> staff = abjad.Staff(r"c'4\p d'8 | e'4.\p | c'4\p d'8\f")
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \p
            d'8
            e'4.
            \p
            c'4
            \p
            d'8
            \f
        }
        >>> auxjad.remove_repeated_dynamics(staff)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \p
            d'8
            e'4.
            c'4
            d'8
            \f
        }

    ..  container:: example

        The input container can also handle subcontainers:

        >>> staff = abjad.Staff([abjad.Note("c'2"),
        ...                      abjad.Chord("<d' f'>2"),
        ...                      abjad.Tuplet((2, 3), "g2 a2 b2"),
        ...                      ])
        >>> abjad.attach(abjad.Dynamic('ppp'), staff[0])
        >>> abjad.attach(abjad.Dynamic('ppp'), staff[1])
        >>> abjad.attach(abjad.Dynamic('ppp'), staff[2][0])
        >>> abjad.f(staff)
        \new Staff
        {
            c'2
            \ppp
            <d' f'>2
            \ppp
            \times 2/3 {
                g2
                \ppp
                a2
                b2
            }
        }
        >>> auxjad.remove_repeated_dynamics(staff)
        >>> abjad.f(staff)
        \new Staff
        {
            c'2
            \ppp
            <d' f'>2
            \times 2/3 {
                g2
                a2
                b2
            }
        }

    ..  container:: example

        By default, repeated dynamics with hairpins in between are not removed,
        but consecutive ones will.

        >>> staff = abjad.Staff(r"c'4\pp\< d'8\f\> | c'4\f d'8\f")
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \pp
            \<
            d'8
            \f
            \>
            c'4
            \f
            d'8
            \f
        }
        >>> auxjad.remove_repeated_dynamics(staff)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \pp
            \<
            d'8
            \f
            \>
            c'4
            \f
            d'8
        }

        To override the previous behaviour, set ``ignore_hairpins=True`` and
        hairpins will be ignored.

        >>> staff = abjad.Staff(r"c'4\pp\< d'8\f\> | c'4\f d'8\f")
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \pp
            \<
            d'8
            \f
            \>
            c'4
            \f
            d'8
            \f
        }
        >>> auxjad.remove_repeated_dynamics(staff, ignore_hairpins=True)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \pp
            \<
            d'8
            \f
            \>
            c'4
            d'8
        }

    ..  container:: example

        By default, rests are treated just like any other leaf and thus notes
        with an identical dynamic separated by an arbitrary number of rests
        will be considered as repeated and the second dynamic will be removed.

        >>> staff = abjad.Staff(r"c'4\pp r2. | c'4\pp")
        >>> auxjad.remove_repeated_dynamics(staff)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \pp
            r2.
            c'4
        }

        To override the previous behaviour, set ``reset_after_rests=True`` and
        dynamics will always be restated after a rest.

        >>> staff = abjad.Staff(r"c'4\pp r2. | c'4\pp")
        >>> auxjad.remove_repeated_dynamics(staff, reset_after_rests=True)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \pp
            r2.
            c'4
            \pp
        }

    ..  container:: example

        The argument ``reset_after_rests`` takes not only boolean values but
        also duration (``abjad.Duration``, tuple, float, etc.). This sets the
        maximum length of rests before which identical dynamics are restated.
        If the total length of rests falls below that value, then repeated
        dynamics are removed.

        In the case below, a rest of ``r2``. is shorter than a duration of
        (4, 4), so the repeated dynamic is removed.

        >>> staff = abjad.Staff(r"c'4\pp r2. | c'4\pp")
        >>> auxjad.remove_repeated_dynamics(staff, reset_after_rests=(4, 4))
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \pp
            r2.
            c'4
        }

        But setting the duration to 2/4 forces the dynamic to be restated.

        >>> staff = abjad.Staff(r"c'4\pp r2. | c'4\pp")
        >>> auxjad.remove_repeated_dynamics(staff, reset_after_rests=2/4)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \pp
            r2.
            c'4
            \pp
        }

    ..  container:: example

        The function also handles measure rests with ``reset_after_rests``.

        >>> staff = abjad.Staff(r"c'4\pp r2. | R1 | c'4\pp")
        >>> auxjad.remove_repeated_dynamics(
        ...     staff,
        ...     reset_after_rests=abjad.Duration(4, 4),
        ... )
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            \pp
            r2.
            R1
            c'4
            \pp
        }
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("'container' must be 'abjad.Container' or child class")
    if not isinstance(ignore_hairpins, bool):
        raise TypeError("'ignore_hairpins' must be 'bool'")
    if not isinstance(reset_after_rests,
                      (bool, int, float, tuple, str, abjad.Duration),
                      ):
        raise TypeError("'reset_after_rests' must be a number, 'bool' or "
                        "'abjad.Duration'")

    leaves = [leaf for leaf in abjad.select(container).leaves()]

    previous_dynamic = None
    duration_since_last_note = abjad.Duration(0)
    for leaf in leaves:
        if type(leaf) in (abjad.Rest, abjad.MultimeasureRest):
            if isinstance(reset_after_rests, bool) and reset_after_rests:
                previous_dynamic = None
            elif reset_after_rests:
                duration_since_last_note += leaf.written_duration
                if duration_since_last_note >= \
                        abjad.Duration(reset_after_rests):
                    previous_dynamic = None
        else:
            duration_since_last_note = abjad.Duration(0)
            indicators = abjad.inspect(leaf).indicators()
            hairpin_present = any([type(indicator) == abjad.StartHairpin for
                                   indicator in indicators])
            for indicator in indicators:
                if type(indicator) == abjad.Dynamic:
                    if hairpin_present and not ignore_hairpins:
                        current_dynamic = None
                    else:
                        current_dynamic = indicator
            if current_dynamic != previous_dynamic:
                previous_dynamic = current_dynamic
            elif current_dynamic is not None:
                abjad.detach(abjad.Dynamic, leaf)



import random


class TenneysContainer():
    r"""TenneysContainer in an implementation of the Dissonant Counterpoint
    Algorithm by James Tenney. This class can be used to randomly select
    elements from an input list, giving more weight to elements which have not
    been selected in recent iterations. In other words, Tenney's algorithm
    uses feedback in order to lower the weight of recently selected elements.

    This implementation is based on the paper: Polansky, L., A. Barnett, and
    M. Winter (2011). `A Few More Words About James Tenney: Dissonant
    Counterpoint and Statistical Feedback'. In: Journal of Mathematics and
    Music 5(2). pp. 63--82.

    ..  container:: example

        The container should be initialised with a list of objects. The
        contents of the list can be absolutely anything.

        >>> container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'])
        >>> container.contents
        ['A', 'B', 'C', 'D', 'E', 'F']

        Applying the ``len()`` function to the container will give the length
        of the container.

        >>> len(container)
        6

        When no other keyword arguments are used, the default probabilities of
        each element in the list is 1.0. Probabilities are not normalised. Use
        the previous_index attribute to check the previously selected index
        (default is None).

        >>> container.probabilities
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> container.previous_index
        None

        Calling the container will output one of its elements, selected
        according to the current probability values. After each call, the
        object updates all probability values, setting the previously selected
        element's probability at 0.0 and raising all other probabilities
        according to a growth function (more on this below).

        >>> result = ''
        >>> for _ in range(30):
        ...     result += container()
        >>> result
        'EDFACEABAFDCEDAFADCBFEDABEDFEC'

        From the result above it is possible to see that there are no immediate
        repetitions of elements (since once selected, their probability is
        always set to 0.0 and will take at least one iteration to grow to a
        non-zero value). Checking the probabilities and previous_index
        attributes will give us their current values.

        >>> container.probabilities
        [6.0, 5.0, 0.0, 3.0, 1.0, 2.0]
        >>> container.previous_index
        2

    ..  container:: example

        This class can take two optional keywords argument during its
        instantiation, namely ``weights`` and ``curvature``. ``weights`` takes
        a list of floats with the individual weights of each element; by
        default,  all weights are set to 1.0. These weights affects the
        effective probability of each element. The other argument,
        ``curvature``, is the exponent of the growth function for all elements.
        The growth function takes as input the number of iterations since an
        element has been last selected, and raise this number by the curvature
        value. If ``curvature`` is set to 1.0 (which is its default value), the
        growth is linear with each iteration. If set to a value larger than 0.0
        and less than 1.0, the growth is negative (or concave), so that the
        chances of an element which is not being selected will grow at ever
        smaller rates as the number of iterations it has not been selected
        increase. If the ``curvature`` is set to 1.0, the growth is linear with
        the number of iterations. If the ``curvature`` is larger than 1.0, the
        curvature is positive (or convex) and the growth will accelerate as the
        number of iterations an element has not been selected grows. Setting
        the curvature to 0.0 will result in an static probability vector with
        all values set to 1.0, except for the previously selected one which
        will be set to 0.0; this will result in a uniformly random selection
        without repetition.

        With linear curvature (default value of 1.0):

        >>> container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'])
        >>> container.curvature
        1.0
        >>> container.weights
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> container.probabilities
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> container()
        'B'
        >>> container.curvature
        1.0
        >>> container.weights
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> container.probabilities
        [2.0, 0.0, 2.0, 2.0, 2.0, 2.0]

    ..  container:: example

        Using a convex curvature:

        >>> container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'],
        ...                                     curvature=0.2,
        ...                                     )
        >>> container.curvature
        0.2
        >>> container.weights
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> container.probabilities
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> container()
        'C'
        >>> container.curvature
        0.2
        >>> container.weights
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> container.probabilities
        [1.148698354997035, 1.148698354997035, 0.0, 1.148698354997035,
        1.148698354997035, 1.148698354997035]

        With a convex curvature, the growth of the probability of each
        non-selected term gets smaller as the number of times it is not
        selected increases. The smaller the curvature is, the less difference
        there will be between any non-previously selected elements. This
        results in sequences which have more chances of a same element being
        near each other. In the sequence below, note how there are many cases
        of a same element being separated only by a single other one, such as
        ``'ACA'`` in index 6.

        >>> result = ''
        >>> for _ in range(30):
        ...     result += container()
        >>> result
        'DACBEDFACABDACECBEFAEDBAFBABFD'

        Checking the probability values at this point outputs:

        >>> container.probabilities
        [1.2457309396155174, 1.148698354997035, 1.6952182030724354, 0.0,
        1.5518455739153598, 1.0]

        As we can see, all non-zero values are relatively close to each other,
        which is why there is a high chance of an element being selected again
        just two iterations apart.

    ..  container:: example

        Using a concave curvature:

        >>> container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'],
        ...                                     curvature=15.2,
        ...                                     )
        >>> container.curvature
        0.2
        >>> container.weights
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> container.probabilities
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> container()
        'C'
        >>> container.curvature
        0.2
        >>> container.weights
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        >>> container.probabilities
        [37640.547696542824, 37640.547696542824, 37640.547696542824, 0.0,
        37640.547696542824, 37640.547696542824]

        With a concave curvature, the growth of the probability of each
        non-selected term gets larger as the number of times it is not selected
        increases. The larger the curvature is, the larger difference there
        will be between any non-previously selected elements. This results in
        sequences which have less chances of a same element being near each
        other. In the sequence below, with a curvature of 15.2, note how the
        elements are as far apart from each other, resulting in a repeating
        string of ``'DFAECB'``.

        >>> result = ''
        >>> for _ in range(30):
        ...     result += container()
        >>> result
        'DFAECBDFAECBDFAECBDFAECBDFAECB'

        Checking the probability values at this point outputs:

        >>> container.probabilities
        [17874877.39956566, 0.0, 1.0, 42106007735.02238,
        37640.547696542824, 1416810830.8957152]

        As we can see, the non-zero values vary wildly. The higher the
        curvature, the higher the difference between these values, making some
        of them much more likely to be selected.

    ..  container:: example

        Each element can also have a fixed weight to themselves. This will
        affect the probability calculation. The example below uses the default
        linear curvature.

        >>> container = auxjad.TenneysContainer(
        ...     ['A', 'B', 'C', 'D', 'E', 'F'],
        ...     weights=[1.0, 1.0, 5.0, 5.0, 10.0, 20.0],
        >>> )
        >>> container.weights
        [1.0, 1.0, 5.0, 5.0, 10.0, 20.0]
        >>> container.probabilities
        [1.0, 1.0, 5.0, 5.0, 10.0, 20.0]
        >>> result = ''
        >>> for _ in range(30):
        ...     result += container()
        >>> result
        'FBEFECFDEADFEDFEDBFECDAFCEDCFE'
        >>> container.weights
        [1.0, 1.0, 5.0, 5.0, 10.0, 20.0]
        >>> container.probabilities
        [7.0, 12.0, 10.0, 15.0, 0.0, 20.0]

    ..  container:: example

        To reset the probability to its initial value, use the method
        ``reset_probabilities()``.

        >>> container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'])
        >>> for _ in range(30):
        ...     container()
        >>> container.probabilities
        [4.0, 3.0, 1.0, 0.0, 5.0, 2.0]
        >>> container.reset_probabilities()
        >>> container.probabilities
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]

    ..  container:: example

        To replace an element in the container, use the method
        ``replace_element()``. This will not affect the current probability
        vector, and the new element will have the same probability as the one
        it replaced.

        >>> container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'])
        >>> for _ in range(30):
        ...     container()
        >>> container.replace_element('foo', 2)
        >>> container.contents
        ['A', 'B', 'foo', 'D', 'E', 'F']
        >>> container.probabilities
        [3.0, 2.0, 1.0, 7.0, 5.0, 0.0]

    ..  container:: example

        A new container of an arbitrary length can be set at any point using
        the method ``set_container()``. Do notice that the probabilities will
        be reset at that point. This method can take the optional keyword
        argument weights similarly to when instantiating the class.

        >>> container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'])
        >>> for _ in range(30):
        ...     container()
        >>> container.probabilities
        [2.0, 1.0, 4.0, 3.0, 0.0, 5.0]
        >>> container.set_container([2, 4, 6, 8])
        >>> container.contents
        [2, 4, 6, 8]
        >>> len(container)
        4
        >>> container.weights
        [1.0, 1.0, 1.0, 1.0]
        >>> container.probabilities
        [1.0, 1.0, 1.0, 1.0]

    ..  container:: example

        To change the curvature value at any point, use the ``set_curvature()``
        method.

        >>> container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'])
        >>> container.curvature
        1.0
        >>> container.set_curvature(0.25)
        >>> container.curvature
        0.25
    """

    def __init__(self,
                 container: list,
                 *,
                 weights: list = None,
                 curvature: float = 1.0,
                 ):
        if not isinstance(container, list):
            raise TypeError("'container' must be 'list'")
        if weights:
            if not isinstance(weights, list):
                raise TypeError("'weights' must be 'list'")
            if not len(container) == len(weights):
                raise ValueError("'weights' must have the same length "
                                 "as 'container'")
            if not all(isinstance(weight, (int, float))
                       for weight in weights):
                raise TypeError("'weights' elements must be "
                                "'int' or 'float'")
        if not isinstance(curvature, float):
            raise TypeError("'curvature' must be 'float'")
        if curvature < 0.0:
            raise ValueError("'curvature' must be larger than 0.0")

        self.contents = container[:]
        if weights:
            self.weights = weights[:]
        else:
            self.weights = [1.0 for _ in self.contents]
        self._counter = [1 for _ in self.contents]
        self.previous_index = None
        self.curvature = curvature
        self._generate_probabilities()

    def __call__(self):
        self.previous_index = random.choices(
            [n for n in range(self.__len__())],
            weights=self.probabilities,
            )[0]
        self._regenerate_counts()
        self._generate_probabilities()
        return self.contents[self.previous_index]

    def __len__(self):
        return len(self.contents)

    def get_element(self, index: int):
        return self.contents[index]

    def replace_element(self, new_element, index: int):
        self.contents = self.contents[:index] \
                      + [new_element] \
                      + self.contents[index+1:]

    def set_container(self,
                      new_container: list,
                      *,
                      weights: list = None,
                      ):
        if not isinstance(new_container, list):
            raise TypeError("'new_container' must be 'list")
        if weights:
            if not isinstance(weights, list):
                raise TypeError("'weights' must be 'list'")
            if not len(new_container) == len(weights):
                raise ValueError("'weights' must have the same length "
                                 "as 'new_container'")
            if not all(isinstance(weight, (int, float))
                       for weight in weights):
                raise TypeError("'weights' elements must be "
                                "'int' or 'float'")

        self.contents = new_container[:]
        if weights:
            self.weights = weights[:]
        else:
            self.weights = [1.0 for _ in self.contents]
        self._generate_probabilities(reset=True)

    def set_curvature(self,
                      curvature: float,
                      ):
        if not isinstance(curvature, float):
            raise TypeError("'curvature' must be 'float'")
        if curvature < 0.0:
            raise ValueError("'curvature' must be larger than 0.0")
        self.curvature = curvature
        self._generate_probabilities()

    def reset_probabilities(self):
        self._counter = [1 for _ in self.contents]
        self._generate_probabilities()

    def _regenerate_counts(self):
        for i in range(self.__len__()):
            if i == self.previous_index:
                self._counter[i] = 0
            else:
                self._counter[i] += 1

    def _generate_probabilities(self,
                                reset: bool = False,
                                ):
        if reset:
            self._counter = [1 for _ in self.contents]
        self.probabilities = []
        for weight, count in zip(self.weights, self._counter):
            self.probabilities.append(weight * self._growth_function(count))

    def _growth_function(self, count):
        return count ** self.curvature



import abjad
from .is_container_full import is_container_full
from .simplified_time_signature_ratio import simplified_time_signature_ratio
from .underfull_duration import underfull_duration


def close_container(container: abjad.Container):
    r"""Changes the time signature of the last bar of an underfull
    ``abjad.Container`` in order to make it full, if necessary.

    ..  container:: example

        Returns the missing duration of the last bar of any container or child
        class. If no time signature is encountered, it uses LilyPond's
        convention and considers the container as in 4/4.

        >>> container1 = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Container(r"c'4 d'4 e'4")
        >>> container3 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4")
        >>> container4 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4 d'4 e'4 f'4")
        >>> auxjad.close_container(container1)
        >>> auxjad.close_container(container2)
        >>> auxjad.close_container(container3)
        >>> auxjad.close_container(container4)
        >>> abjad.f(container1)
        {
            c'4
            d'4
            e'4
            f'4
        }
        >>> abjad.f(container2)
        {
            %%% \time 3/4 %%%
            c'4
            d'4
            e'4
        }
        >>> abjad.f(container3)
        {
            c'4
            d'4
            e'4
            f'4
            %%% \time 1/4 %%%
            c'4
        }
        >>> abjad.f(container4)
        {
            c'4
            d'4
            e'4
            f'4
            c'4
            d'4
            e'4
            f'4
        }

    ..  container:: example

        Handles any time signatures as well as changes of time signature.

        >>> container1 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4 g'")
        >>> container2 = abjad.Container(r"\time 3/4 a2. \time 2/4 c'4")
        >>> container3 = abjad.Container(r"\time 5/4 g1 ~ g4 \time 4/4 af'2")
        >>> auxjad.close_container(container1)
        >>> auxjad.close_container(container2)
        >>> auxjad.close_container(container3)
        >>> abjad.f(container1)
        {
            %%% \time 4/4 %%%
            c'4
            d'4
            e'4
            f'4
            %%% \time 1/4 %%%
            g'4
        }
        >>> abjad.f(container2)
        {
            %%% \time 3/4 %%%
            a2.
            %%% \time 1/4 %%%
            c'4
        }
        >>> abjad.f(container3)
        {
            %%% \time 5/4 %%%
            g1
            ~
            g4
            %%% \time 2/4 %%%
            af'2
        }

        Notice that the time signatures in the output are commented out with
        ``%%%``. This is because Abjad only applies time signatures to
        containers that belong to a ``abjad.Staff``. The present function works
        with either ``abjad.Container`` and ``abjad.Staff``.

        >>> container = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4 g'")
        >>> auxjad.close_container(container)
        >>> abjad.f(container)
        {
            %%% \time 4/4 %%%
            c'4
            d'4
            e'4
            f'4
            %%% \time 1/4 %%%
            g'4
        }
        >>> staff = abjad.Staff([container])
        >>> abjad.f(container)
        {
            \time 4/4
            c'4
            d'4
            e'4
            f'4
            \time 1/4
            g'4
        }

    ..  container:: example

        Correctly handles partial time signatures.

        >>> container = abjad.Container(r"c'4 d'4 e'4 f'4 g'4")
        >>> time_signature = abjad.TimeSignature((3, 4), partial=(1, 4))
        >>> abjad.attach(time_signature, container[0])
        >>> auxjad.close_container(container)
        >>> abjad.f(container)
        {
            %%% \partial 4 %%%
            %%% \time 3/4 %%%
            c'4
            d'4
            e'4
            f'4
            %%% \time 1/4 %%%
            g'4
        }

    ..  container:: example

        If a container is malformed, i.e. it has an underfilled bar before a
        time signature change, the function raises a ``ValueError`` exception.

        >>> container = abjad.Container(r"\time 5/4 g''1 \time 4/4 f'4")
        >>> auxjad.close_container(container)
        ValueError: 'container' is malformed, with an underfull bar preceeding
        a time signature change
    """
    if not isinstance(container, abjad.Container):
        raise TypeError("'container' must be 'abjad.Container' or child class")
    if not is_container_full(container):
        missing_duration = underfull_duration(container)
        leaves = abjad.select(container).leaves()
        for leaf in leaves[::-1]:
            time_signature = abjad.inspect(leaf).effective(abjad.TimeSignature)
            if time_signature:
                last_time_signature = time_signature
                break
        else:
            last_time_signature = abjad.TimeSignature((4, 4))
        last_bar_duration = last_time_signature.duration - missing_duration
        final_bar_time_signature = abjad.TimeSignature(last_bar_duration)
        final_bar_time_signature = simplified_time_signature_ratio(
            final_bar_time_signature)
        duration = 0
        for leaf in leaves[::-1]:
            duration += abjad.inspect(leaf).duration()
            if duration == last_bar_duration:
                if abjad.inspect(leaf).indicators(abjad.TimeSignature):
                    abjad.detach(abjad.TimeSignature, leaf)
                abjad.attach(final_bar_time_signature, leaf)
                break



import copy
import random
import abjad
from .are_leaves_tieable import are_leaves_tieable


class _LoopWindowGeneric():
    r"""This is the parent class of all LoopWindowXxx classes. It implements
    all common methods and attributes, and initialises those using its set_
    methods.
    """

    def __init__(self,
                 head_position,
                 window_size,
                 step_size,
                 max_steps,
                 repetition_chance,
                 forward_bias,
                 ):
        self._first_window = True
        self.set_head_position(head_position)
        self.set_window_size(window_size)
        self.set_step_size(step_size)
        self.set_max_steps(max_steps)
        self.set_repetition_chance(repetition_chance)
        self.set_forward_bias(forward_bias)

    def __call__(self) -> abjad.Selection:
        self._move_head()
        if self._done():
            raise RuntimeError("'container' has been exhausted")
        self._slice_container()
        return copy.deepcopy(self._current_window)

    def __iter__(self):
        return self

    def __next__(self) -> abjad.Selection:
        self._move_head()
        if self._done():
            raise StopIteration
        self._slice_container()
        return copy.deepcopy(self._current_window)

    def set_window_size(self,
                        window_size: int,
                        ):
        if not isinstance(window_size, int):
            raise TypeError("'window_size' must be 'int'")
        if window_size < 1:
            raise ValueError("'window_size' must be greater than zero")
        if window_size > self._container.__len__():
            raise ValueError("'window_size' must be smaller than or equal to "
                             "the length of 'container'")
        self.window_size = window_size

    def set_step_size(self,
                      step_size: int,
                      ):
        if not isinstance(step_size, int):
            raise TypeError("'step_size' must be 'int'")
        if step_size < 1:
            raise ValueError("'step_size' must be greater than zero")
        if step_size >= self._container.__len__():
            raise ValueError("'step_size' must be smaller than the length of "
                             "'container'")
        self.step_size = step_size

    def set_max_steps(self,
                      max_steps: int,
                      ):
        if not isinstance(max_steps, int):
            raise TypeError("'max_steps' must be 'int'")
        if max_steps < 1:
            raise ValueError("'max_steps' must be greater than zero")
        self.max_steps = max_steps

    def set_repetition_chance(self,
                              repetition_chance: float,
                              ):
        if not isinstance(repetition_chance, float):
            raise TypeError("'repetition_chance' must be 'float'")
        if repetition_chance < 0.0 or repetition_chance > 1.0:
            raise ValueError("'repetition_chance' must be between 0.0 and 1.0")
        self.repetition_chance = repetition_chance

    def set_forward_bias(self,
                         forward_bias: float,
                         ):
        if not isinstance(forward_bias, float):
            raise TypeError("'forward_bias' must be 'float'")
        if forward_bias < 0.0 or forward_bias > 1.0:
            raise ValueError("'forward_bias' must be between 0.0 and 1.0")
        self.forward_bias = forward_bias

    def set_head_position(self,
                          head_position: int,
                          ):
        if not isinstance(head_position, int):
            raise TypeError("'head_position' must be 'int'")
        if head_position < 0:
            raise ValueError("'head_position' must be a positive 'int'")
        if head_position >= self._container.__len__():
            raise ValueError("'head_position' must be smaller than the length "
                             "of 'container'")
        self.head_position = head_position

    def get_current_window(self) -> (list, abjad.Selection):
        return copy.deepcopy(self._current_window)

    def output_all(self,
                   *,
                   tie_identical_pitches: bool = False,
                   ) -> abjad.Selection:
        dummy_container = abjad.Container()
        while True:
            try:
                if not tie_identical_pitches or len(dummy_container) == 0:
                    dummy_container.append(self.__call__())
                else:
                    new_window = self.__call__()
                    leaf1 = abjad.select(new_window).leaves()[0]
                    leaf2 = abjad.select(dummy_container).leaves()[-1]
                    if are_leaves_tieable(leaf1, leaf2):
                        abjad.attach(abjad.Tie(), dummy_container[-1])
                    dummy_container.append(new_window)
            except:
                break
        result = dummy_container[:]
        dummy_container[:] = []
        return result

    def output_n(self,
                 n: int,
                 *,
                 tie_identical_pitches: bool = False,
                 ) -> abjad.Selection:
        dummy_container = abjad.Container()
        for _ in range(n):
            if not tie_identical_pitches or len(dummy_container) == 0:
                dummy_container.append(self.__call__())
            else:
                new_window = self.__call__()
                leaf1 = abjad.select(new_window).leaves()[0]
                leaf2 = abjad.select(dummy_container).leaves()[-1]
                if are_leaves_tieable(leaf1, leaf2):
                    abjad.attach(abjad.Tie(), dummy_container[-1])
                dummy_container.append(new_window)
        result = dummy_container[:]
        dummy_container[:] = []
        return result

    def _move_head(self):
        if not self._first_window:  # first window always at initial position
            if self.repetition_chance == 0.0 \
                    or random.random() > self.repetition_chance:
                step = self.step_size * random.randint(1, self.max_steps)
                diretion = self._biased_choice(self.forward_bias)
                self.head_position += step * diretion
        else:
            self._first_window = False

    def _done(self) -> bool:
        return self.head_position >= self._container.__len__() or \
            self.head_position < 0

    def _slice_container(self):
        pass

    @staticmethod
    def _biased_choice(bias):
        return random.choices([1, -1], weights=[bias, 1.0-bias])[0]



import random
import abjad
import auxjad


def test_TenneysContainer_01():
    random.seed(43714)
    container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'])
    assert container.contents == ['A', 'B', 'C', 'D', 'E', 'F']
    assert container.curvature == 1.0
    assert container.weights == [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    assert container.probabilities == [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    assert container.previous_index is None
    assert len(container) == 6
    result = ''
    for _ in range(30):
        result += container()
    assert result == 'BDAFCBEFCDFDBAEDFCABDEABCDEBFE'
    assert container.previous_index == 4
    assert container.weights == [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
    assert container.probabilities == [7.0, 2.0, 5.0, 4.0, 0.0, 1.0]


def test_TenneysContainer_02():
    random.seed(37248)
    container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'],
                                        curvature=0.001,
                                        )
    result = ''
    for _ in range(30):
        result += container()
    assert result == 'CDCBFECBCEBCFCEDFACDBADCADCDFE'
    assert container.probabilities == [1.0016107337527294,
                                       1.002199640244188,
                                       1.001099215984204,
                                       1.0006933874625807,
                                       0.0,
                                       1.0,
                                       ]


def test_TenneysContainer_03():
    random.seed(76231)
    container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'],
                                        curvature=15.2,
                                        )
    result = ''
    for _ in range(30):
        result += container()
    assert result == 'DFAECBDFAECBDFAECBDFAECBDFAECB'
    assert container.probabilities == [17874877.39956566,
                                       0.0,
                                       1.0,
                                       42106007735.02238,
                                       37640.547696542824,
                                       1416810830.8957152,
                                       ]


def test_TenneysContainer_04():
    random.seed(14625)
    container = auxjad.TenneysContainer(
        ['A', 'B', 'C', 'D', 'E', 'F'],
        weights=[1.0, 1.0, 5.0, 5.0, 10.0, 20.0],
    )
    assert container.weights == [1.0, 1.0, 5.0, 5.0, 10.0, 20.0]
    assert container.probabilities == [1.0, 1.0, 5.0, 5.0, 10.0, 20.0]
    result = ''
    for _ in range(30):
        result += container()
    assert result == 'FBEFECFDEADFEDFEDBFECDAFCEDCFE'
    assert container.weights == [1.0, 1.0, 5.0, 5.0, 10.0, 20.0]
    assert container.probabilities == [7.0, 12.0, 10.0, 15.0, 0.0, 20.0]


def test_TenneysContainer_05():
    random.seed(21169)
    container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'])
    for _ in range(30):
        container()
    container.replace_element('foo', 2)
    assert container.contents == ['A', 'B', 'foo', 'D', 'E', 'F']
    assert container.probabilities == [3.0, 2.0, 1.0, 7.0, 5.0, 0.0]


def test_TenneysContainer_06():
    random.seed(54267)
    container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'])
    for _ in range(30):
        container()
    assert container.probabilities == [2.0, 1.0, 4.0, 3.0, 0.0, 5.0]
    container.set_container([2, 4, 6, 8])
    assert container.contents == [2, 4, 6, 8]
    assert len(container) == 4
    assert container.weights == [1.0, 1.0, 1.0, 1.0]
    assert container.probabilities == [1.0, 1.0, 1.0, 1.0]


def test_TenneysContainer_07():
    container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'])
    assert container.curvature == 1.0
    container.set_curvature(0.25)
    assert container.curvature == 0.25


def test_TenneysContainer_08():
    random.seed(99651)
    container = auxjad.TenneysContainer(['A', 'B', 'C', 'D', 'E', 'F'])
    for _ in range(30):
        container()
    assert container.probabilities == [4.0, 3.0, 1.0, 0.0, 5.0, 2.0]
    container.reset_probabilities()
    assert container.probabilities == [1.0, 1.0, 1.0, 1.0, 1.0, 1.0]



import abjad
import pytest
import auxjad


def test_underfull_duration_01():
    container1 = abjad.Container(r"c'4 d'4 e'4 f'4")
    container2 = abjad.Container(r"c'4 d'4 e'4")
    container3 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4")
    container4 = abjad.Container(r"c'4 d'4 e'4 f'4 | c'4 d'4 e'4 f'4")
    assert auxjad.underfull_duration(container1) == 0
    assert auxjad.underfull_duration(container2) == 1/4
    assert auxjad.underfull_duration(container3) == 3/4
    assert auxjad.underfull_duration(container4) == 0


def test_underfull_duration_02():
    container1 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4")
    container2 = abjad.Container(r"\time 3/4 a2. \time 2/4 r2")
    container3 = abjad.Container(r"\time 5/4 g1 ~ g4 \time 4/4 af'2")
    container4 = abjad.Container(r"\time 6/8 c'2 ~ c'8")
    assert auxjad.underfull_duration(container1) == 0
    assert auxjad.underfull_duration(container2) == 0
    assert auxjad.underfull_duration(container3) == 1/2
    assert auxjad.underfull_duration(container4) == 1/8


def test_underfull_duration_03():
    container = abjad.Container(r"c'4 d'4 e'4 f'4")
    time_signature = abjad.TimeSignature((3, 4), partial=(1, 4))
    abjad.attach(time_signature, container[0])
    assert auxjad.underfull_duration(container) == 0


def test_underfull_duration_04():
    container1 = abjad.Container(r"R1")
    container2 = abjad.Container(r"\time 3/4 R1*3/4 \time 2/4 r2")
    container3 = abjad.Container(r"\time 5/4 R1*5/4 \time 4/4 g''4")
    container4 = abjad.Container(r"\time 6/8 R1*1/2")
    assert auxjad.underfull_duration(container1) == 0
    assert auxjad.underfull_duration(container2) == 0
    assert auxjad.underfull_duration(container3) == 3/4
    assert auxjad.underfull_duration(container4) == 1/4


def test_underfull_duration_05():
    container = abjad.Container(r"\time 5/4 g''1 \time 4/4 f'1")
    with pytest.raises(ValueError):
        assert auxjad.underfull_duration(container)



import abjad
import auxjad


def test_are_containers_equal_01():
    container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
    container2 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
    assert auxjad.are_containers_equal(container1, container2)


def test_are_containers_equal_02():
    container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
    container2 = abjad.Staff(r"\times 3/2 {c'4 d'4 e'4} f'4 <g' a'>2 r2")
    assert not auxjad.are_containers_equal(container1, container2)


def test_are_containers_equal_03():
    container1 = abjad.Staff(r"c'4\pp d'4 e'4-. f'4 <g' a'>2-> r2")
    container2 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
    assert auxjad.are_containers_equal(container1, container2)


def test_are_containers_equal_04():
    container1 = abjad.Staff(r"c'4\pp d'4 e'4-. f'4 <g' a'>2-> r2")
    container2 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
    assert not auxjad.are_containers_equal(container1,
                                           container2,
                                           include_indicators=True,
                                           )


def test_are_containers_equal_05():
    container1 = abjad.Staff(r"c'4\pp d'4 e'4-. f'4 <g' a'>2-> r2")
    container2 = abjad.Staff(r"c'4\pp d'4 e'4-. f'4 <g' a'>2-> r2")
    assert auxjad.are_containers_equal(container1,
                                       container2,
                                       include_indicators=True,
                                       )


def test_are_containers_equal_06():
    container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
    container2 = abjad.Staff(r"c'4 \grace{d'4} e'4 f'4 <g' a'>2 r2")
    assert not auxjad.are_containers_equal(container1, container2)


def test_are_containers_equal_07():
    container1 = abjad.Staff(r"c'4 d'4 e'4 f'4 <g' a'>2 r2")
    container2 = abjad.Staff(r"c'4 \grace{c''4} d'4 e'4 f'4 <g' a'>2 r2")
    assert not auxjad.are_containers_equal(container1, container2)


def test_are_containers_equal_08():
    container1 = abjad.Staff(r"c'4 \grace{c''16} d'4 e'4 f'4 <g' a'>2 r2")
    container2 = abjad.Staff(r"c'4 \grace{c''16} d'4 e'4 f'4 <g' a'>2 r2")
    assert auxjad.are_containers_equal(container1, container2)


def test_are_containers_equal_09():
    container1 = abjad.Staff(r"c'4 \grace{c''4} d'4 e'4 f'4 <g' a'>2 r2")
    container2 = abjad.Staff(r"c'4 \grace{c''8} d'4 e'4 f'4 <g' a'>2 r2")
    assert not auxjad.are_containers_equal(container1, container2)


def test_are_containers_equal_10():
    container1 = abjad.Staff(r"c'4 \grace{c''4} d'4 e'4 f'4 <g' a'>2 r2")
    container2 = abjad.Staff(r"c'4 \grace{b4} d'4 e'4 f'4 <g' a'>2 r2")
    assert not auxjad.are_containers_equal(container1, container2)



import copy
import abjad
from ._LoopWindowGeneric import _LoopWindowGeneric
from .simplified_time_signature_ratio import simplified_time_signature_ratio


class LoopWindowByElements(_LoopWindowGeneric):
    r"""Takes an ``abjad.Container`` as input as well as an integer
    representing the number of elements per looping window, then outputs a
    container with the elements processed in the looping process. For instance,
    if the initial container had the leaves ``[A, B, C, D, E, F]`` and the
    looping window was size three, the output would be:

    ``A B C B C D C D E D E F E F F``

    This can be better visualised as:

    .. code-block:: none

        A B C
          B C D
            C D E
              D E F
                E F
                  F

    ..  container:: example

        Usage is similar to other factory classes. It takes a container (or
        child class equivalent) and the number of elements of the window as
        arguments. Each call of the object, in this case ``looper()``, will
        move the window forwards and output the result.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopWindowByElements(input_music,
        ...                                      window_size=3,
        ...                                      )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 11/8
            d'2
            e'4
            f'2
            ~
            f'8
        }

        The method ``get_current_window()`` will output the current window
        without moving the head forwards.

        >>> notes = looper.get_current_window()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 11/8
            d'2
            e'4
            f'2
            ~
            f'8
        }

    ..  container:: example

        The instances of ``LoopWindowByElements`` can also be used as an
        iterator, which can then be used in a for loop to exhaust all windows.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4")
        >>> looper = auxjad.LoopWindowByElements(input_music,
        ...                                      window_size=2,
        ...                                      )
        >>> for window in looper:
        ...     staff = abjad.Staff(window)
        ...     abjad.f(staff)
        \new Staff
        {
            \time 3/4
            c'4
            d'2
        }
        \new Staff
        {
            \time 3/4
            d'2
            e'4
        }
        \new Staff
        {
            \time 1/4
            e'4
        }

    ..  container:: example

        This class can take many optional keyword arguments during its
        creation. ``step_size`` dictates the size of each individual step in
        number of elements (default value is 1). ``max_steps`` sets the maximum
        number of steps that the window can advance when the object is called,
        ranging between 1 and the input value (default is also 1).
        ``repetition_chance`` sets the chance of a window result repeating
        itself (that is, the window not moving forwards when called). It should
        range from 0.0 to 1.0 (default 0.0, i.e. no repetition).
        ``forward_bias`` sets the chance of the window moving forward instead
        of backwards. It should range from 0.0 to 1.0 (default 1.0, which means
        the window can only move forwards. A value of 0.5 gives 50% chance of
        moving forwards while a value of 0.0 will move the window only
        backwards). Finally, ``head_position`` can be used to offset the
        starting position of the looping window. It must be an integer and its
        default value is 0.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopWindowByElements(input_music,
        ...                                      window_size=3,
        ...                                      step_size=1,
        ...                                      max_steps=2,
        ...                                      repetition_chance=0.25,
        ...                                      forward_bias=0.2,
        ...                                      head_position=0,
        ...                                      )
        >>> looper.window_size
        3
        >>> looper.step_size
        1
        >>> looper.repetition_chance
        0.25
        >>> looper.forward_bias
        0.2
        >>> looper.max_steps
        2
        >>> looper.head_position
        0

        Use the ``set_`` methods below to change these values after
        initialisation.

        >>> looper.set_window_size(2)
        >>> looper.set_step_size(2)
        >>> looper.set_max_steps(3)
        >>> looper.set_repetition_chance(0.1)
        >>> looper.set_forward_bias(0.8)
        >>> looper.set_head_position(2)
        >>> looper.window_size
        2
        >>> looper.step_size
        2
        >>> looper.max_steps
        3
        >>> looper.repetition_chance
        0.1
        >>> looper.forward_bias
        0.8
        >>> looper.head_position
        2

    ..  container:: example

        To disable time signatures altogether, initialise
        ``LoopWindowByElements`` with the keyword argument
        ``omit_time_signature`` set to ``True`` (default is ``False``), or use
        the ``set_omit_time_signature()`` method after initialisation.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopWindowByElements(input_music,
        ...                                      window_size=3,
        ...                                      omit_time_signature=True,
        ...                                      )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            c'4
            d'2
            e'4
        }

    ..  container:: example

        The function ``len()`` can be used to get the total number of elements
        in the container.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopWindowByElements(input_music,
        ...                                      window_size=3,
        ...                                      )
        >>> len(looper)
        5

    ..  container:: example

        To run through the whole process and output it as a single container,
        from the initial head position until the process outputs the single
        last element, use the method ``output_all()``.

        >>> input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LoopWindowByElements(input_music,
        ...                                      window_size=2,
        ...                                      )
        >>> window = looper.output_all()
        >>> staff = abjad.Staff(window)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 2/4
            c'4
            d'4
            \time 2/4
            d'4
            e'4
            \time 2/4
            e'4
            f'4
            \time 1/4
            f'4
        }

    ..  container:: example

        When using ``output_all()``, set the keyword argument
        ``tie_identical_pitches`` to ``True`` in order to tie identical notes
        or chords at the end and beginning of consecutive windows.

        >>> input_music = abjad.Container(r"c'4 d'2 r8 d'4 <e' g'>8 r4 f'2. "
        ...                               "<e' g'>16")
        >>> looper = auxjad.LoopWindowByElements(input_music,
        ...                                      window_size=4,
        ...                                      )
        >>> music = looper.output_all(tie_identical_pitches=True)
        >>> staff = abjad.Staff(music)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 9/8
            c'4
            d'2
            r8
            d'4
            ~
            \time 4/4
            d'2
            r8
            d'4
            <e' g'>8
            \time 3/4
            r8
            d'4
            <e' g'>8
            r4
            \time 11/8
            d'4
            <e' g'>8
            r4
            f'2.
            \time 19/16
            <e' g'>8
            r4
            f'2.
            <e' g'>16
            \time 17/16
            r4
            f'2.
            <e' g'>16
            \time 13/16
            f'2.
            <e' g'>16
            ~
            \time 1/16
            <e' g'>16
        }

    ..  container:: example

        To run through just part of the process and output it as a single
        container, starting from the initial head position, use the method
        ``output_n()`` and pass the number of iterations as argument. Similarly
        to ``output_all()``, the keyword argument ``tie_identical_pitches`` is
        available for tying pitches.

        >>> input_music = abjad.Container(r"c'4 d'4 e'4 f'4")
        >>> looper = auxjad.LoopWindowByElements(input_music,
        ...                                      window_size=2,
        ...                                      )
        >>> window = looper.output_n(2)
        >>> staff = abjad.Staff(window)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 2/4
            c'4
            d'4
            \time 2/4
            d'4
            e'4
        }

    .. container:: example

        To change the size of the window after instantiation, use the method
        ``set_window_size()``. In the example below, the initial window is of
        size 3, and so the first call of the looper object outputs the first,
        second, and third leaves. The window size is then set to 4, and the
        looper is called again, moving to the leaf in the next position, thus
        outputting the second, third, fourth, and fifth leaves.

        >>> input_music = abjad.Container(r"c'4 d'2 e'4 f'2 ~ f'8 g'1")
        >>> looper = auxjad.LoopWindowByElements(input_music,
        ...                                      window_size=3,
        ...                                      )
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 4/4
            c'4
            d'2
            e'4
        }
        >>> looper.set_window_size(4)
        >>> notes = looper()
        >>> staff = abjad.Staff(notes)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 19/8
            d'2
            e'4
            f'2
            ~
            f'8
            g'1
        }

    ..  container:: example

        This class can handle tuplets, but the output is not ideal and so this
        functionality should be considered experimental. Time signatures will
        be correct when dealing with partial tuplets (thus having non-standard
        values in their denominators), but each individual note of a tuplet
        will have the ratio printed above it.

        >>> input_music = abjad.Container(r"c'4 d'8 \times 2/3 {a4 g2}")
        >>> looper = auxjad.LoopWindowByElements(input_music,
        ...                                      window_size=2,
        ...                                      )
        >>> window = looper.output_all()
        >>> staff = abjad.Staff(window)
        >>> abjad.f(staff)
        \new Staff
        {
            \time 3/8
            c'4
            d'8
            #(ly:expect-warning "strange time signature found")
            \time 7/24
            d'8
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                a4
            }
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                \time 2/4
                a4
            }
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                g2
            }
            \tweak edge-height #'(0.7 . 0)
            \times 2/3 {
                #(ly:expect-warning "strange time signature found")
                \time 2/6
                g2
            }
        }
    """

    def __init__(self,
                 container: abjad.Container,
                 *,
                 window_size: int,
                 step_size: int = 1,
                 max_steps: int = 1,
                 repetition_chance: float = 0.0,
                 forward_bias: float = 1.0,
                 head_position: int = 0,
                 omit_time_signature: bool = False,
                 ):
        if not isinstance(container, abjad.Container):
            raise TypeError("'container' must be 'abjad.Container' or child "
                            "class")
        self._container = abjad.select(container).logical_ties()
        self.set_omit_time_signature(omit_time_signature)
        super().__init__(head_position,
                         window_size,
                         step_size,
                         max_steps,
                         repetition_chance,
                         forward_bias,
                         )


    def __len__(self) -> int:
        return len(self._container)

    def set_omit_time_signature(self,
                                omit_time_signature: bool,
                                ):
        if not isinstance(omit_time_signature, bool):
            raise TypeError("'omit_time_signature' must be 'bool'")
        self.omit_time_signature = omit_time_signature

    def _slice_container(self) -> abjad.Selection:
        start = self.head_position
        end = self.head_position + self.window_size
        logical_ties = self._container[start:end]
        dummy_container = abjad.Container()
        time_signature_duration = 0
        for logical_tie in logical_ties:
            effective_duration = abjad.inspect(logical_tie).duration()
            logical_tie_ = copy.deepcopy(logical_tie)
            dummy_container.append(logical_tie_)
            multiplier = effective_duration / logical_tie_.written_duration
            logical_tie_ = abjad.mutate(logical_tie_).scale(multiplier)
            time_signature_duration += effective_duration
        if len(logical_ties) > 0 and not self.omit_time_signature:
            time_signature = abjad.TimeSignature(time_signature_duration)
            time_signature = simplified_time_signature_ratio(time_signature)
            abjad.attach(time_signature,
                         abjad.select(dummy_container).leaves()[0],
                         )
        self._current_window = dummy_container[:]
        dummy_container[:] = []



import abjad
import pytest
import auxjad


def test_LoopWindowByList_01():
    input_list = ['A', 'B', 'C', 'D', 'E', 'F']
    looper = auxjad.LoopWindowByList(input_list, window_size=3)
    assert looper() == ['A', 'B', 'C']
    assert looper() == ['B', 'C', 'D']
    assert looper.get_current_window() == ['B', 'C', 'D']


def test_LoopWindowByList_02():
    input_list = ['A', 'B', 'C', 'D', 'E', 'F']
    looper = auxjad.LoopWindowByList(input_list,
                                     window_size=3,
                                     step_size=1,
                                     max_steps=2,
                                     repetition_chance=0.25,
                                     forward_bias=0.2,
                                     head_position=0,
                                     )
    assert looper.window_size == 3
    assert looper.step_size == 1
    assert looper.max_steps == 2
    assert looper.repetition_chance == 0.25
    assert looper.forward_bias == 0.2
    assert looper.head_position == 0
    looper.set_window_size(2)
    looper.set_step_size(2)
    looper.set_max_steps(3)
    looper.set_repetition_chance(0.1)
    looper.set_forward_bias(0.8)
    looper.set_head_position(2)
    assert looper.window_size == 2
    assert looper.step_size == 2
    assert looper.max_steps == 3
    assert looper.repetition_chance == 0.1
    assert looper.forward_bias == 0.8
    assert looper.head_position == 2


def test_LoopWindowByList_03():
    input_list = ['A', 'B', 'C', 'D', 'E', 'F']
    looper = auxjad.LoopWindowByList(input_list, window_size=3)
    assert looper.head_position == 0
    looper()
    assert looper.head_position == 0
    looper()
    assert looper.head_position == 1
    looper()
    assert looper.head_position == 2


def test_LoopWindowByList_04():
    input_list = ['A', 'B', 'C', 'D', 'E', 'F']
    looper = auxjad.LoopWindowByList(input_list, window_size=3)
    assert len(looper) == 6


def test_LoopWindowByList_05():
    input_list = ['A', 'B', 'C', 'D']
    looper = auxjad.LoopWindowByList(input_list, window_size=3)
    assert looper.output_all() == ['A', 'B', 'C', 'B', 'C', 'D', 'C', 'D', 'D']


def test_LoopWindowByList_06():
    input_list = ['A', 'B', 'C', 'D', 'E', 'F']
    looper = auxjad.LoopWindowByList(input_list, window_size=3)
    assert looper() == ['A', 'B', 'C']
    looper.set_window_size(4)
    assert looper() == ['B', 'C', 'D', 'E']


def test_LoopWindowByList_07():
    input_list = [123, 'foo', (3, 4), 3.14]
    looper = auxjad.LoopWindowByList(input_list, window_size=3)
    assert looper() == [123, 'foo', (3, 4)]


def test_LoopWindowByList_08():
    input_list = [
        abjad.Container(r"c'4 d'4 e'4 f'4"),
        abjad.Container(r"fs'1"),
        abjad.Container(r"r2 bf2"),
        abjad.Container(r"c''2. r4"),
    ]
    looper = auxjad.LoopWindowByList(input_list, window_size=3)
    staff = abjad.Staff()
    for element in looper.output_all():
        staff.append(element)
    assert format(staff) == abjad.String.normalize(
        r'''
        \new Staff
        {
            {
                c'4
                d'4
                e'4
                f'4
            }
            {
                fs'1
            }
            {
                r2
                bf2
            }
            {
                fs'1
            }
            {
                r2
                bf2
            }
            {
                c''2.
                r4
            }
            {
                r2
                bf2
            }
            {
                c''2.
                r4
            }
            {
                c''2.
                r4
            }
        }
        ''')


def test_LoopWindowByList_09():
    input_list = ['A', 'B', 'C', 'D']
    looper = auxjad.LoopWindowByList(input_list, window_size=3)
    assert looper.__next__() == ['A', 'B', 'C']
    assert looper.__next__() == ['B', 'C', 'D']
    assert looper.__next__() == ['C', 'D']
    assert looper.__next__() == ['D']
    with pytest.raises(StopIteration):
        assert looper.__next__()



def test_LoopWindowByList_10():
    wrong_type_input = 'foo'
    input_list = ['A', 'B', 'C', 'D']
    with pytest.raises(TypeError):
        assert auxjad.LoopWindowByList(wrong_type_input, window_size=3)
        assert auxjad.LoopWindowByList(input_list, window_size='foobar')
        assert auxjad.LoopWindowByList(input_list,
                                       window_size=3,
                                       step_size='foobar',
                                       )
        assert auxjad.LoopWindowByList(input_list,
                                       window_size=3,
                                       max_steps='foobar',
                                       )
        assert auxjad.LoopWindowByList(input_list,
                                       window_size=3,
                                       repetition_chance='foobar',
                                       )
        assert auxjad.LoopWindowByList(input_list,
                                       window_size=3,
                                       head_position='foobar',
                                       )
    with pytest.raises(ValueError):
        assert auxjad.LoopWindowByList(input_list, window_size=-1)
        assert auxjad.LoopWindowByList(input_list,
                                       window_size=3,
                                       step_size=-1)
        assert auxjad.LoopWindowByList(input_list,
                                       window_size=3,
                                       step_size=100)
        assert auxjad.LoopWindowByList(input_list,
                                       window_size=3,
                                       max_steps=-1)
        assert auxjad.LoopWindowByList(input_list,
                                       window_size=3,
                                       repetition_chance=-0.3)
        assert auxjad.LoopWindowByList(input_list,
                                       window_size=3,
                                       repetition_chance=1.4)
        assert auxjad.LoopWindowByList(input_list,
                                       window_size=3,
                                       head_position=-1)
        assert auxjad.LoopWindowByList(input_list,
                                       window_size=3,
                                       head_position=100)


def test_LoopWindowByList_11():
    input_list = ['A', 'B', 'C', 'D']
    looper = auxjad.LoopWindowByList(input_list, window_size=3)
    assert looper.output_n(2) == ['A', 'B', 'C', 'B', 'C', 'D']


def test_LoopWindowByList_12():
    input_list = ['A', 'B', 'C', 'D']
    looper = auxjad.LoopWindowByList(input_list, window_size=3)
    with pytest.raises(RuntimeError):
        looper.output_n(100)


def test_LoopWindowByList_13():
    input_list = ['A', 'B', 'C', 'D']
    looper = auxjad.LoopWindowByList(input_list,
                                     window_size=2,
                                     head_position=2,
                                     forward_bias=0.0,
                                     )
    assert looper.output_all() == ['C', 'D', 'B', 'C', 'A', 'B']


def test_LoopWindowByList_14():
    input_list = ['A', 'B', 'C', 'D']
    looper = auxjad.LoopWindowByList(input_list,
                                     window_size=2,
                                     head_position=0,
                                     forward_bias=0.0,
                                     )
    assert looper.output_all() == ['A', 'B']



import abjad
from .close_container import close_container
from .is_container_full import is_container_full
from .simplified_time_signature_ratio import simplified_time_signature_ratio
from .underfull_duration import underfull_duration


def sync_containers(*containers: abjad.Container,
                    use_multimeasure_rests: bool = True,
                    adjust_last_time_signature: bool = True,
                    ):
    r"""Takes an arbitrary number of ``abjad.Container``'s (or child classes),
    finds the longest one and adds rests to all the shorter ones, making them
    the same length. By default, it rewrites the last time signature if
    necessary, and uses multi-measure rests whenever possible.

    ..  container:: example

        Input two or more containers. This function will fill the shortest ones
        with rests ensuring all their lengths become the same.

        >>> container1 = abjad.Container(r"\time 4/4 g'2.")
        >>> container2 = abjad.Container(r"\time 4/4 c'1")
        >>> auxjad.sync_containers(container1, container2)
        >>> abjad.f(container1)
        {
            %%% \time 4/4 %%%
            g'2.
            r4
        }
        >>> abjad.f(container2)
        {
            %%% \time 4/4 %%%
            c'1
        }

        Notice that the time signatures in the output are commented out with
        ``%%%``. This is because Abjad only applies time signatures to
        containers that belong to a ``abjad.Staff``. The present function works
        with either ``abjad.Container`` and ``abjad.Staff``.

        >>> container1 = abjad.Container(r"\time 4/4 g'2.")
        >>> container2 = abjad.Container(r"\time 4/4 c'1")
        >>> auxjad.sync_containers(container1, container2)
        >>> abjad.f(container1)
        {
            %%% \time 4/4 %%%
            g'2.
            r4
        }
        >>> staff = abjad.Staff([container1])
        >>> abjad.f(container1)
        {
            \time 4/4
            g'2.
            r4
        }

    ..  container:: example

        If all containers have the same size, no modification is applied.

        >>> container1 = abjad.Container(r"\time 3/4 g'2.")
        >>> container2 = abjad.Container(r"\time 3/4 c'2.")
        >>> auxjad.sync_containers(container1, container2)
        >>> abjad.f(container1)
        {
            %%% \time 3/4 %%%
            g'2.
        }
        >>> abjad.f(container2)
        {
            %%% \time 3/4 %%%
            c'2.
        }

    ..  container:: example

        By default, this function closes the longest container by rewriting the
        time signature of its last bar if necessary (if it is underfull), and
        uses multi-measure rests whenever possible.

        >>> container1 = abjad.Container(r"\time 4/4 g'1 | f'4")
        >>> container2 = abjad.Container(r"\time 4/4 c'1")
        >>> auxjad.sync_containers(container1, container2)
        >>> abjad.f(container1)
        {
            %%% \time 4/4 %%%
            g'1
            %%% \time 1/4 %%%
            f'4
        }
        >>> abjad.f(container2)
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 1/4 %%%
            R1*1/4
        }

    ..  container:: example

        To disable multi-measure rests, set the keyword argument
        ``use_multimeasure_rests`` to ``False``.

        >>> container1 = abjad.Container(r"\time 4/4 g'1 | f'4")
        >>> container2 = abjad.Container(r"\time 4/4 c'1")
        >>> auxjad.sync_containers(container1,
        ...                        container2,
        ...                        use_multimeasure_rests=False,
        ...                        )
        >>> abjad.f(container1)
        {
            %%% \time 4/4 %%%
            g'1
            %%% \time 1/4 %%%
            f'4
        }
        >>> abjad.f(container2)
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 1/4 %%%
            r4
        }

    ..  container:: example

        To allow containers to be left open (with underfull bars), set the
        keyword argument ``adjust_last_time_signature`` to ``False``.

        >>> container1 = abjad.Container(r"\time 4/4 g'1 | f'4")
        >>> container2 = abjad.Container(r"\time 4/4 c'1")
        >>> auxjad.sync_containers(container1,
        ...                        container2,
        ...                        adjust_last_time_signature=False,
        ...                        )
        >>> abjad.f(container1)
        {
            %%% \time 4/4 %%%
            g'1
            f'4
        }
        >>> abjad.f(container2)
        {
            %%% \time 4/4 %%%
            c'1
            r4
        }

    ..  container:: example

        This function can take an arbitrary number of containers.

        >>> container1 = abjad.Container(r"\time 4/4 c'1 | g'4")
        >>> container2 = abjad.Container(r"\time 4/4 c'1 | g'2")
        >>> container3 = abjad.Container(r"\time 4/4 c'1 | g'2.")
        >>> container4 = abjad.Container(r"\time 4/4 c'1")
        >>> auxjad.sync_containers(container1,
        ...                        container2,
        ...                        container3,
        ...                        container4,
        ...                        )
        >>> abjad.f(container1)
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 3/4 %%%
            g'4
            r2
        }
        >>> abjad.f(container2)
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 3/4 %%%
            g'2
            r4
        }
        >>> abjad.f(container3)
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 3/4 %%%
            g'2.
        }
        >>> abjad.f(container4)
        {
            %%% \time 4/4 %%%
            c'1
            %%% \time 3/4 %%%
            R1*3/4
        }

    ..  container:: example

        The containers can be of different length, can have different time
        signatures, and can contain time signature changes as well.

        >>> container1 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Container(r"\time 3/4 a2. \time 4/4 c'4")
        >>> container3 = abjad.Container(r"\time 5/4 g''1 ~ g''4")
        >>> container4 = abjad.Container(r"\time 6/8 c'2")
        >>> auxjad.sync_containers(container1,
        ...                        container2,
        ...                        container3,
        ...                        container4,
        ...                        )
        >>> abjad.f(container1)
        {
            %%% \time 4/4 %%%
            c'4
            d'4
            e'4
            f'4
            %%% \time 1/4 %%%
            R1*1/4
        }
        >>> abjad.f(container2)
        {
            %%% \time 3/4 %%%
            a2.
            %%% \time 2/4 %%%
            c'4
            r4
        }
        >>> abjad.f(container3)
        {
            %%% \time 5/4 %%%
            g''1
            ~
            g''4
        }
        >>> abjad.f(container4)
        {
            %%% \time 6/8 %%%
            c'2
            r4
            %%% \time 2/4 %%%
            R1*1/2
        }

        It's important to note that LilyPond does not support simultanoues
        staves with different time signatures (i.e. polymetric notation) by
        default. In order to enable it, the ``"Timing_translator"`` and
        ``"Default_bar_line_engraver"`` must be removed from the ``Score``
        context and added to the ``Staff`` context. Below is a full example of
        how this can be accomplished using Abjad.

        >>> container1 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4")
        >>> container2 = abjad.Container(r"\time 3/4 a2. \time 4/4 c'4")
        >>> container3 = abjad.Container(r"\time 5/4 g''1 ~ g''4")
        >>> container4 = abjad.Container(r"\time 6/8 c'2")
        >>> auxjad.sync_containers(container1,
        ...                        container2,
        ...                        container3,
        ...                        container4,
        ...                        )
        >>> staves = [abjad.Staff([container1]),
        ...           abjad.Staff([container2]),
        ...           abjad.Staff([container3]),
        ...           abjad.Staff([container4]),
        ...           ]
        >>> score = abjad.Score(staves)
        >>> lilypond_file = abjad.LilyPondFile.new()
        >>> score_block = abjad.Block(name='score')
        >>> layout_block = abjad.Block(name='layout')
        >>> score_block.items.append(score)
        >>> score_block.items.append(layout_block)
        >>> lilypond_file.items.append(score_block)
        >>> layout_block.items.append(
        ...     r'''
        ...     \context {
        ...         \Score
        ...         \remove "Timing_translator"
        ...         \remove "Default_bar_line_engraver"
        ...     }
        ...     \context {
        ...         \Staff
        ...         \consists "Timing_translator"
        ...         \consists "Default_bar_line_engraver"
        ...     }
        ...     ''')
        >>> abjad.show(lilypond_file)

    ..  container:: example

        If one or more containers is malformed, i.e. it has an underfilled bar
        before a time signature change, the function raises a ``ValueError``
        exception.

        >>> container1 = abjad.Container(r"\time 4/4 g'1 | f'4")
        >>> container2 = abjad.Container(r"\time 5/4 c'1 | \time 4/4 d'4")
        >>> auxjad.sync_containers(container1, container2)
        ValueError: at least one 'container' is malformed, with an underfull
        bar preceeding a time signature change
    """
    for container in containers:
        if not isinstance(container, abjad.Container):
            raise TypeError("all 'containers' must be 'abjad.Container' or "
                            "child class")
        try:
            is_container_full(container)
        except:
            raise ValueError("at least one 'container' is malformed, with an "
                             "underfull bar preceeding a time signature "
                             "change")
    durations = [abjad.inspect(container[:]).duration() for container
                 in containers]
    max_duration = max(durations)
    for container, duration in zip(containers, durations):
        duration_difference = max_duration - duration
        if duration_difference > abjad.Duration(0):
            # handling duration left in the last bar, if any
            if not is_container_full(container):
                duration_left = underfull_duration(container)
                underfull_rests_duration = min(duration_difference,
                                               duration_left,
                                               )
                underfull_rests = abjad.LeafMaker()(None,
                                                    underfull_rests_duration,
                                                    )
                duration_difference -= underfull_rests_duration
                container.extend(underfull_rests)
                if adjust_last_time_signature and duration_difference == \
                        abjad.Duration(0):
                    close_container(container)
                if duration_difference == abjad.Duration(0):
                    continue
            # finding out last effective time signature
            for leaf in abjad.select(container).leaves()[::-1]:
                effective_time_signature = abjad.inspect(leaf).effective(
                    abjad.TimeSignature)
                if effective_time_signature:
                    break
            else:
                effective_time_signature = abjad.TimeSignature((4, 4))
            # creating new bars for any leftover duratin
            if use_multimeasure_rests and (adjust_last_time_signature or \
                    duration_difference == effective_time_signature.duration):
                multiplier = duration_difference.pair
                if multiplier[0] == multiplier[1]:  # avoiding R1 * 1
                    multiplier = None
                rests = [abjad.MultimeasureRest((4, 4), multiplier=multiplier)]
            else:
                rests = abjad.LeafMaker()(None, duration_difference)
            if adjust_last_time_signature:
                rests_time_signature = abjad.TimeSignature(duration_difference)
                rests_time_signature = simplified_time_signature_ratio(
                    rests_time_signature)
                if rests_time_signature != effective_time_signature:
                    abjad.attach(rests_time_signature, rests[0])
            container.extend(rests)
        else:
            # closing longest container if necessary
            if adjust_last_time_signature and not is_container_full(container):
                close_container(container)


# SETUP
container1 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4")
# container2 = abjad.Container(r"\time 3/4 a2. c'4")
# container3 = abjad.Container(r"\time 5/4 g''1 ~ g''4")
# container4 = abjad.Container(r"\time 6/8 c'2")
container2 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4 g' a'")
container3 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4 g' a' b'")
container4 = abjad.Container(r"\time 4/4 c'4 d'4 e'4 f'4 g'")

staves = [abjad.Staff([container1]),
          abjad.Staff([container2]),
          abjad.Staff([container3]),
          abjad.Staff([container4]),
          ]

sync_containers(container1,
                container2,
                container3,
                container4,
                #use_multimeasure_rests=False,
                )

# =============================================================================
# TODO: function for this below?
lilypond_file = abjad.LilyPondFile.new()
score = abjad.Score(staves)
score_block = abjad.Block(name='score')
layout_block = abjad.Block(name='layout')
score_block.items.append(score)
score_block.items.append(layout_block)
lilypond_file.items.append(score_block)
layout_block.items.append(
    r'''
    \context {
        \Score
        \remove "Timing_translator"
        \remove "Default_bar_line_engraver"
    }
    \context {
        \Staff
        \consists "Timing_translator"
        \consists "Default_bar_line_engraver"
    }
    ''')

# abjad.show(lilypond_file)
